import typing
import types
import os
import signal

from .. import devices
from .. import worker


class SerialExecutor:
    def __init__(
        self,
        worker: worker.Worker,
    ) -> None:
        self.worker = worker

        self.currently_working = False

        has_soft_timeout = self.worker.config.timeouts.soft_timeout > 0
        has_hard_timeout = self.worker.config.timeouts.hard_timeout > 0
        has_critical_timeout = self.worker.config.timeouts.critical_timeout > 0
        should_use_a_killer = has_soft_timeout or has_hard_timeout or has_critical_timeout
        if should_use_a_killer:
            self.killer = devices.process_killer.ProcessKiller(
                pid_to_kill=os.getpid(),
                sleep_interval=0.1,
                soft_timeout=self.worker.config.timeouts.soft_timeout,
                hard_timeout=self.worker.config.timeouts.hard_timeout,
                critical_timeout=self.worker.config.timeouts.critical_timeout,
            )

            self.original_int = signal.getsignal(signal.SIGINT)
            self.original_abrt = signal.getsignal(signal.SIGABRT)
            signal.signal(signal.SIGABRT, self.sigabrt_handler)
            signal.signal(signal.SIGINT, self.sigint_handler)
        else:
            self.killer = None

    def sigabrt_handler(
        self,
        signal_num: int,
        frame: types.FrameType,
    ) -> None:
        if self.currently_working:
            raise worker.WorkerHardTimedout()

    def sigint_handler(
        self,
        signal_num: int,
        frame: types.FrameType,
    ) -> None:
        if self.currently_working:
            raise worker.WorkerSoftTimedout()

    def execute_tasks(
        self,
        tasks: typing.Iterable[typing.Dict[str, typing.Any]],
    ) -> None:
        for task in tasks:
            self.execute_task(
                task=task,
            )

    def execute_task(
        self,
        task: typing.Dict[str, typing.Any],
    ) -> None:
        try:
            self.pre_work(
                task=task,
            )

            returned_value = self.worker.work(
                task=task,
            )

            self.post_work(
                task=task,
                success=True,
            )

            self.worker._on_success(
                task=task,
                returned_value=returned_value,
            )
        except Exception as exception:
            self.post_work(
                task=task,
                success=False,
                exception=exception,
            )

            if isinstance(exception, worker.WorkerTimedout):
                self.worker._on_timeout(
                    task=task,
                )
            elif isinstance(exception, worker.WorkerRetry):
                self.worker._on_retry(
                    task=task,
                )
            elif isinstance(exception, worker.WorkerMaxRetries):
                self.worker._on_max_retries(
                    task=task,
                )
            elif isinstance(exception, worker.WorkerRequeue):
                self.worker._on_requeue(
                    task=task,
                )
            else:
                self.worker._on_failure(
                    task=task,
                    exception=exception,
                )

    def pre_work(
        self,
        task: typing.Dict[str, typing.Any],
    ) -> None:
        try:
            self.worker.pre_work(
                task=task,
            )
        except Exception as exception:
            self.worker.logger.error(
                msg=f'pre_work has failed: {exception}',
                extra={
                    'task': task,
                },
            )

        self.currently_working = True

        if self.killer:
            self.killer.start()

    def post_work(
        self,
        task: typing.Dict[str, typing.Any],
        success: bool,
        exception: typing.Optional[Exception] = None,
    ) -> None:
        if self.killer:
            self.killer.stop_and_reset()

        self.currently_working = False

        try:
            self.worker.post_work(
                task=task,
                success=success,
                exception=exception,
            )
        except Exception as exception:
            self.worker.logger.error(
                msg=f'post_work has failed: {exception}',
                extra={
                    'task': task,
                },
            )

    def __del__(
        self,
    ) -> None:
        if self.killer:
            try:
                self.killer.kill()

                signal.signal(signal.SIGABRT, self.original_abrt)
                signal.signal(signal.SIGINT, self.original_int)
            except Exception:
                pass



import typing
import msgpack


class Serializer:
    name = 'msgpack'

    msgpack_packer = msgpack.Packer(
        autoreset=True,
    )

    @staticmethod
    def serialize(
        data: typing.Any,
    ) -> bytes:
        return Serializer.msgpack_packer.pack(data)

    @staticmethod
    def unserialize(
        data: bytes,
    ) -> typing.Any:
        return msgpack.unpackb(
            packed=data,
            strict_map_key=False,
        )



import unittest
import unittest.mock
import datetime
import threading
import time

import sergeant.task_queue
import sergeant.connector
import sergeant.encoder


class TaskQueueTestCase:
    order_matters = True

    def test_purge_tasks(
        self,
    ):
        for test_task_queue in self.test_task_queues:
            test_task_queue.purge_tasks(
                task_name='test_task',
            )
            self.assertEqual(
                first=test_task_queue.number_of_enqueued_tasks(
                    task_name='test_task',
                ),
                second=0,
            )
            task = test_task_queue.craft_task(
                task_name='test_task',
            )
            test_task_queue.apply_async_one(
                task=task,
                priority='NORMAL',
            )
            self.assertEqual(
                first=test_task_queue.number_of_enqueued_tasks(
                    task_name='test_task',
                ),
                second=1,
            )
            test_task_queue.purge_tasks(
                task_name='test_task',
            )
            self.assertEqual(
                first=test_task_queue.number_of_enqueued_tasks(
                    task_name='test_task',
                ),
                second=0,
            )

    def test_number_of_enqueued_tasks(
        self,
    ):
        for test_task_queue in self.test_task_queues:
            test_task_queue.purge_tasks(
                task_name='test_task',
            )
            self.assertEqual(
                first=test_task_queue.number_of_enqueued_tasks(
                    task_name='test_task',
                ),
                second=0,
            )
            task = test_task_queue.craft_task(
                task_name='test_task',
            )
            test_task_queue.apply_async_one(
                task=task,
                priority='NORMAL',
            )
            self.assertEqual(
                first=test_task_queue.number_of_enqueued_tasks(
                    task_name='test_task',
                ),
                second=1,
            )
            test_task_queue.purge_tasks(
                task_name='test_task',
            )
            self.assertEqual(
                first=test_task_queue.number_of_enqueued_tasks(
                    task_name='test_task',
                ),
                second=0,
            )

            test_task_queue.apply_async_many(
                task_name='test_task',
                tasks=[task] * 100,
                priority='NORMAL',
            )
            self.assertEqual(
                first=test_task_queue.number_of_enqueued_tasks(
                    task_name='test_task',
                ),
                second=100,
            )
            test_task_queue.apply_async_many(
                task_name='test_task',
                tasks=[task] * 1000,
                priority='NORMAL',
            )
            self.assertEqual(
                first=test_task_queue.number_of_enqueued_tasks(
                    task_name='test_task',
                ),
                second=1100,
            )
            test_task_queue.purge_tasks(
                task_name='test_task',
            )
            self.assertEqual(
                first=test_task_queue.number_of_enqueued_tasks(
                    task_name='test_task',
                ),
                second=0,
            )

    def test_craft_task(
        self,
    ):
        for test_task_queue in self.test_task_queues:
            task = test_task_queue.craft_task(
                task_name='test_task',
                kwargs={},
            )
            current_date = datetime.datetime.utcnow().timestamp()
            date = task.pop('date')
            self.assertAlmostEqual(
                first=date / (10 ** 8),
                second=current_date / (10 ** 8),
            )
            self.assertEqual(
                first=task,
                second={
                    'name': 'test_task',
                    'kwargs': {},
                    'run_count': 0,
                }
            )

            task = test_task_queue.craft_task(
                task_name='test_task',
                kwargs={
                    'a': 1,
                    'b': 2,
                },
            )
            current_date = datetime.datetime.utcnow().timestamp()
            date = task.pop('date')
            self.assertAlmostEqual(
                first=date / (10 ** 8),
                second=current_date / (10 ** 8),
            )
            self.assertEqual(
                first=task,
                second={
                    'name': 'test_task',
                    'kwargs': {
                        'a': 1,
                        'b': 2,
                    },
                    'run_count': 0,
                }
            )

    def test_wait_queue_empty(
        self,
    ):
        for test_task_queue in self.test_task_queues:
            test_task_queue.purge_tasks(
                task_name='test_task',
            )
            task = test_task_queue.craft_task(
                task_name='test_task',
                kwargs={},
            )
            test_task_queue.apply_async_one(
                task=task,
                priority='NORMAL',
            )
            purge_tasks_timer = threading.Timer(
                interval=1.0,
                function=test_task_queue.purge_tasks,
                args=(
                    'test_task',
                ),
            )
            purge_tasks_timer.start()

            before = time.time()
            test_task_queue.wait_queue_empty(
                task_name='test_task',
                sample_interval=0.1,
            )
            after = time.time()
            self.assertTrue(
                expr=1.2 > after - before > 1.0,
            )

    def test_apply_async_one(
        self,
    ):
        for test_task_queue in self.test_task_queues:
            test_task_queue.purge_tasks(
                task_name='test_task',
            )
            task_one = test_task_queue.craft_task(
                task_name='test_task',
                kwargs={
                    'arg': 'one',
                },
            )
            task_two = test_task_queue.craft_task(
                task_name='test_task',
                kwargs={
                    'arg': 'two',
                },
            )
            task_three = test_task_queue.craft_task(
                task_name='test_task',
                kwargs={},
            )

            test_task_queue.apply_async_one(
                task=task_one,
                priority='NORMAL',
            )
            test_task_queue.apply_async_one(
                task=task_two,
                priority='NORMAL',
            )
            test_task_queue.apply_async_one(
                task=task_three,
                priority='NORMAL',
            )
            task_one_test = test_task_queue.get_tasks(
                task_name='test_task',
                number_of_tasks=1,
            )[0]
            task_two_test = test_task_queue.get_tasks(
                task_name='test_task',
                number_of_tasks=1,
            )[0]
            task_three_test = test_task_queue.get_tasks(
                task_name='test_task',
                number_of_tasks=1,
            )[0]
            if self.order_matters:
                self.assertEqual(
                    first=task_one,
                    second=task_one_test,
                )
                self.assertEqual(
                    first=task_two,
                    second=task_two_test,
                )
                self.assertEqual(
                    first=task_three,
                    second=task_three_test,
                )
            else:
                self.assertIn(
                    member=task_one,
                    container=[
                        task_one_test,
                        task_two_test,
                        task_three_test,
                    ],
                )
                self.assertIn(
                    member=task_two,
                    container=[
                        task_one_test,
                        task_two_test,
                        task_three_test,
                    ],
                )
                self.assertIn(
                    member=task_three,
                    container=[
                        task_one_test,
                        task_two_test,
                        task_three_test,
                    ],
                )

    def test_apply_async_many(
        self,
    ):
        for test_task_queue in self.test_task_queues:
            test_task_queue.purge_tasks(
                task_name='test_task_one',
            )
            task_one = test_task_queue.craft_task(
                task_name='test_task_one',
                kwargs={
                    'arg': 'one',
                },
            )
            task_two = test_task_queue.craft_task(
                task_name='test_task_one',
                kwargs={
                    'arg': 'two',
                },
            )
            test_task_queue.apply_async_many(
                task_name='test_task_one',
                tasks=[
                    task_one,
                    task_two,
                ],
                priority='NORMAL',
            )
            task_one_test = test_task_queue.get_tasks(
                task_name='test_task_one',
                number_of_tasks=1,
            )[0]
            task_two_test = test_task_queue.get_tasks(
                task_name='test_task_one',
                number_of_tasks=1,
            )[0]

            if self.order_matters:
                self.assertEqual(
                    first=task_one,
                    second=task_one_test,
                )
                self.assertEqual(
                    first=task_two,
                    second=task_two_test,
                )
            else:
                self.assertIn(
                    member=task_one,
                    container=[
                        task_one_test,
                        task_two_test,
                    ],
                )
                self.assertIn(
                    member=task_two,
                    container=[
                        task_one_test,
                        task_two_test,
                    ],
                )

            self.assertEqual(
                first=task_one,
                second=task_one_test,
            )
            self.assertEqual(
                first=task_two,
                second=task_two_test,
            )

    def test_queue_priority(
        self,
    ):
        for test_task_queue in self.test_task_queues:
            test_task_queue.purge_tasks(
                task_name='test_task',
            )
            task_NORMAL_priority = test_task_queue.craft_task(
                task_name='test_task',
                kwargs={
                    'priority': 'NORMAL',
                },
            )
            task_HIGH_priority = test_task_queue.craft_task(
                task_name='test_task',
                kwargs={
                    'priority': 'HIGH',
                },
            )
            test_task_queue.apply_async_one(
                task=task_NORMAL_priority,
                priority='NORMAL',
            )
            test_task_queue.apply_async_one(
                task=task_NORMAL_priority,
                priority='NORMAL',
            )
            test_task_queue.apply_async_one(
                task=task_HIGH_priority,
                priority='HIGH',
            )
            test_task_queue.apply_async_one(
                task=task_HIGH_priority,
                priority='HIGH',
            )
            test_task_queue.apply_async_one(
                task=task_NORMAL_priority,
                priority='NORMAL',
            )
            test_task_queue.apply_async_one(
                task=task_NORMAL_priority,
                priority='NORMAL',
            )
            test_task_queue.apply_async_one(
                task=task_HIGH_priority,
                priority='HIGH',
            )
            test_task_queue.apply_async_one(
                task=task_HIGH_priority,
                priority='HIGH',
            )
            self.assertEqual(
                first=test_task_queue.number_of_enqueued_tasks(
                    task_name='test_task',
                ),
                second=8,
            )

            high_priority_tasks = test_task_queue.get_tasks(
                task_name='test_task',
                number_of_tasks=2,
            )
            high_priority_tasks += test_task_queue.get_tasks(
                task_name='test_task',
                number_of_tasks=2,
            )
            low_priority_tasks = test_task_queue.get_tasks(
                task_name='test_task',
                number_of_tasks=2,
            )
            low_priority_tasks += test_task_queue.get_tasks(
                task_name='test_task',
                number_of_tasks=2,
            )
            low_priority_tasks += test_task_queue.get_tasks(
                task_name='test_task',
                number_of_tasks=2,
            )
            self.assertEqual(
                first=[task_HIGH_priority['kwargs']['priority']] * 4,
                second=[task['kwargs']['priority'] for task in high_priority_tasks],
            )
            self.assertEqual(
                first=[task_NORMAL_priority['kwargs']['priority']] * 4,
                second=[task['kwargs']['priority'] for task in low_priority_tasks],
            )

    def test_get_tasks(
        self,
    ):
        for test_task_queue in self.test_task_queues:
            test_task_queue.purge_tasks(
                task_name='test_task_one',
            )
            test_task_queue.purge_tasks(
                task_name='test_task_two',
            )
            task_one = test_task_queue.craft_task(
                task_name='test_task_one',
                kwargs={
                    'arg': 'one',
                },
            )
            task_two = test_task_queue.craft_task(
                task_name='test_task_one',
                kwargs={
                    'arg': 'two',
                },
            )
            test_task_queue.apply_async_many(
                task_name='test_task_one',
                tasks=[
                    task_one,
                    task_two,
                ],
                priority='NORMAL',
            )
            tasks_one = test_task_queue.get_tasks(
                task_name='test_task_one',
                number_of_tasks=3,
            )
            self.assertIn(
                member=task_one,
                container=tasks_one,
            )
            self.assertIn(
                member=task_two,
                container=tasks_one,
            )

    def test_retry(
        self,
    ):
        for test_task_queue in self.test_task_queues:
            test_task_queue.purge_tasks(
                task_name='test_task',
            )
            task_one = test_task_queue.craft_task(
                task_name='test_task',
                kwargs={},
            )
            self.assertEqual(task_one['run_count'], 0)
            test_task_queue.apply_async_one(
                task=task_one,
                priority='NORMAL',
            )
            task_one = test_task_queue.get_tasks(
                task_name='test_task',
                number_of_tasks=1,
            )[0]

            test_task_queue.retry(
                task=task_one,
            )
            task_one = test_task_queue.get_tasks(
                task_name='test_task',
                number_of_tasks=1,
            )[0]
            self.assertEqual(
                first=task_one['run_count'],
                second=1,
            )

    def test_requeue(
        self,
    ):
        for test_task_queue in self.test_task_queues:
            test_task_queue.purge_tasks(
                task_name='test_task',
            )
            task_one = test_task_queue.craft_task(
                task_name='test_task',
                kwargs={},
            )
            self.assertEqual(task_one['run_count'], 0)
            test_task_queue.apply_async_one(
                task=task_one,
                priority='NORMAL',
            )
            task_one = test_task_queue.get_tasks(
                task_name='test_task',
                number_of_tasks=1,
            )[0]

            test_task_queue.requeue(
                task=task_one,
            )
            task_one = test_task_queue.get_tasks(
                task_name='test_task',
                number_of_tasks=1,
            )[0]
            self.assertEqual(
                first=task_one['run_count'],
                second=0,
            )

    def test_get_set_key(
        self,
    ):
        for test_task_queue in self.test_task_queues:
            test_task_queue.delete_key(
                name='key_name',
            )

            test_key_value = test_task_queue.get_key(
                name='key_name',
            )
            self.assertIsNone(
                obj=test_key_value,
            )

            key_was_set = test_task_queue.set_key(
                name='key_name',
                value={
                    'key': 'value',
                },
            )
            self.assertTrue(
                expr=key_was_set,
            )

            key_was_set = test_task_queue.set_key(
                name='key_name',
                value={
                    'key': 'value',
                },
            )
            self.assertFalse(
                expr=key_was_set,
            )

            test_key_value = test_task_queue.get_key(
                name='key_name',
            )
            self.assertEqual(
                first=test_key_value,
                second={
                    'key': 'value',
                },
            )

            key_was_deleted = test_task_queue.delete_key(
                name='key_name',
            )
            self.assertTrue(
                expr=key_was_deleted,
            )

            test_key_value = test_task_queue.get_key(
                name='key_name',
            )
            self.assertIsNone(
                obj=test_key_value,
            )


class RedisSingleServerTaskQueueTestCase(
    TaskQueueTestCase,
    unittest.TestCase,
):
    order_matters = True

    def setUp(
        self,
    ):
        self.test_task_queues = []

        connector_obj = sergeant.connector.redis.Connector(
            host='127.0.0.1',
            port=6379,
            password=None,
            database=0,
        )

        compressor_names = list(sergeant.encoder.compressor.__compressors__.keys())
        compressor_names.append(None)
        serializer_names = sergeant.encoder.serializer.__serializers__.keys()
        for compressor_name in compressor_names:
            for serializer_name in serializer_names:
                encoder_obj = sergeant.encoder.encoder.Encoder(
                    compressor_name=compressor_name,
                    serializer_name=serializer_name,
                )
                self.test_task_queues.append(
                    sergeant.task_queue.TaskQueue(
                        connector=connector_obj,
                        encoder=encoder_obj,
                    )
                )


class RedisClusterSingleServerTaskQueueTestCase(
    TaskQueueTestCase,
    unittest.TestCase,
):
    order_matters = True

    def setUp(
        self,
    ):
        self.test_task_queues = []

        connector_obj = sergeant.connector.redis_cluster.Connector(
            nodes=[
                {
                    'host': '127.0.0.1',
                    'port': 6379,
                    'password': None,
                    'database': 0,
                },
            ]
        )

        compressor_names = list(sergeant.encoder.compressor.__compressors__.keys())
        compressor_names.append(None)
        serializer_names = sergeant.encoder.serializer.__serializers__.keys()
        for compressor_name in compressor_names:
            for serializer_name in serializer_names:
                encoder_obj = sergeant.encoder.encoder.Encoder(
                    compressor_name=compressor_name,
                    serializer_name=serializer_name,
                )
                self.test_task_queues.append(
                    sergeant.task_queue.TaskQueue(
                        connector=connector_obj,
                        encoder=encoder_obj,
                    )
                )


class RedisClusterMultipleServerTaskQueueTestCase(
    TaskQueueTestCase,
    unittest.TestCase,
):
    order_matters = False

    def setUp(
        self,
    ):
        self.test_task_queues = []

        connector_obj = sergeant.connector.redis_cluster.Connector(
            nodes=[
                {
                    'host': '127.0.0.1',
                    'port': 6379,
                    'password': None,
                    'database': 0,
                },
                {
                    'host': '127.0.0.1',
                    'port': 6380,
                    'password': None,
                    'database': 0,
                },
            ]
        )

        compressor_names = list(sergeant.encoder.compressor.__compressors__.keys())
        compressor_names.append(None)
        serializer_names = sergeant.encoder.serializer.__serializers__.keys()
        for compressor_name in compressor_names:
            for serializer_name in serializer_names:
                encoder_obj = sergeant.encoder.encoder.Encoder(
                    compressor_name=compressor_name,
                    serializer_name=serializer_name,
                )
                self.test_task_queues.append(
                    sergeant.task_queue.TaskQueue(
                        connector=connector_obj,
                        encoder=encoder_obj,
                    )
                )


class MongoTaskQueueTestCase(
    TaskQueueTestCase,
    unittest.TestCase,
):
    order_matters = False

    def setUp(
        self,
    ):
        self.test_task_queues = []

        connector_obj = sergeant.connector.mongo.Connector(
            mongodb_uri='mongodb://localhost:27017/',
        )

        compressor_names = list(sergeant.encoder.compressor.__compressors__.keys())
        compressor_names.append(None)
        serializer_names = sergeant.encoder.serializer.__serializers__.keys()
        for compressor_name in compressor_names:
            for serializer_name in serializer_names:
                encoder_obj = sergeant.encoder.encoder.Encoder(
                    compressor_name=compressor_name,
                    serializer_name=serializer_name,
                )
                self.test_task_queues.append(
                    sergeant.task_queue.TaskQueue(
                        connector=connector_obj,
                        encoder=encoder_obj,
                    )
                )



import typing

from . import compressor
from . import serializer


class Encoder:
    def __init__(
        self,
        compressor_name: typing.Optional[str],
        serializer_name: str,
    ) -> None:
        if compressor_name:
            self.compressor = compressor.__compressors__[compressor_name]()
        else:
            self.compressor = None

        self.serializer = serializer.__serializers__[serializer_name]()

    def encode(
        self,
        data: typing.Any,
    ) -> bytes:
        serialized_data = self.serializer.serialize(
            data=data,
        )

        if self.compressor:
            serialized_data = self.compressor.compress(
                data=serialized_data,
            )

        return serialized_data

    def decode(
        self,
        data: bytes,
    ) -> typing.Any:
        if self.compressor:
            data = self.compressor.decompress(
                data=data,
            )

        unserialized_data = self.serializer.unserialize(
            data=data,
        )

        return unserialized_data



import redis
import random
import typing


class Connector:
    name = 'redis_cluster'

    def __init__(
        self,
        nodes: typing.List[typing.Dict[str, typing.Any]],
    ) -> None:
        self.nodes = nodes

        self.connections = [
            redis.Redis(
                host=node['host'],
                port=node['port'],
                password=node['password'],
                db=node['database'],
                retry_on_timeout=True,
                socket_keepalive=True,
                socket_connect_timeout=10,
                socket_timeout=60,
            )
            for node in nodes
        ]

        self.master_connection = self.connections[0]

        random.shuffle(self.connections)

    def rotate_connections(
        self,
    ) -> None:
        self.connections = self.connections[1:] + self.connections[:1]

    def key_set(
        self,
        key: str,
        value: bytes,
    ) -> bool:
        is_new = self.master_connection.set(
            name=key,
            value=value,
            nx=True,
        )

        return is_new is True

    def key_get(
        self,
        key: str,
    ) -> bytes:
        return self.master_connection.get(
            name=key,
        )

    def key_delete(
        self,
        key: str,
    ) -> bool:
        return self.master_connection.delete(key) > 0

    def queue_pop(
        self,
        queue_name: str,
    ) -> typing.Optional[bytes]:
        connections = self.connections

        for connection in connections:
            value = connection.lpop(
                name=queue_name,
            )

            self.rotate_connections()

            if value:
                return value

        return None

    def queue_pop_bulk(
        self,
        queue_name: str,
        number_of_items: int,
    ) -> typing.List[bytes]:
        values = []
        connections = self.connections
        current_count = number_of_items

        for connection in connections:
            pipeline = connection.pipeline()

            pipeline.lrange(queue_name, 0, current_count - 1)
            pipeline.ltrim(queue_name, current_count, -1)

            value = pipeline.execute()

            self.rotate_connections()

            values += value[0]

            if len(values) == number_of_items:
                return values

            current_count = number_of_items - len(values)

        return values

    def queue_push(
        self,
        queue_name: str,
        item: bytes,
        priority: str = 'NORMAL',
    ) -> bool:
        if priority == 'HIGH':
            self.connections[0].lpush(queue_name, item)
        else:
            self.connections[0].rpush(queue_name, item)

        self.rotate_connections()

        return True

    def queue_push_bulk(
        self,
        queue_name: str,
        items: typing.Iterable[bytes],
        priority: str = 'NORMAL',
    ) -> bool:
        if priority == 'HIGH':
            self.connections[0].lpush(queue_name, *items)
        else:
            self.connections[0].rpush(queue_name, *items)

        self.rotate_connections()

        return True

    def queue_length(
        self,
        queue_name: str,
    ) -> int:
        total_len = 0

        for connection in self.connections:
            total_len += connection.llen(
                name=queue_name,
            )

        return total_len

    def queue_delete(
        self,
        queue_name: str,
    ) -> bool:
        deleted_count = 0

        for connection in self.connections:
            deleted_count += connection.delete(queue_name)

        return deleted_count > 0



from . import msgpack
from . import json
from . import pickle


__serializers__ = {
    msgpack.Serializer.name: msgpack.Serializer,
    json.Serializer.name: json.Serializer,
    pickle.Serializer.name: pickle.Serializer,
}



import pymongo
import typing


class Connector:
    name = 'mongo'

    def __init__(
        self,
        mongodb_uri: str,
    ) -> None:
        self.mongodb_uri = mongodb_uri
        self.connection = pymongo.MongoClient(
            host=mongodb_uri,
        )

        self.connection.sergeant.task_queue.create_index(
            [
                (
                    'queue_name',
                    pymongo.DESCENDING,
                ),
                (
                    'priority',
                    pymongo.ASCENDING,
                ),
            ],
            background=True,
        )
        self.connection.sergeant.keys.create_index(
            [
                (
                    'value',
                    pymongo.ASCENDING,
                ),
            ],
            background=True,
        )

    def key_set(
        self,
        key: str,
        value: bytes,
    ) -> bool:
        update_one_result = self.connection.sergeant.keys.update_one(
            filter={
                'key': key,
            },
            update={
                '$set': {
                    'key': key,
                    'value': value,
                },
            },
            upsert=True,
        )
        if update_one_result.upserted_id is not None:
            return True
        else:
            return False

    def key_get(
        self,
        key: str,
    ) -> typing.Optional[bytes]:
        document = self.connection.sergeant.keys.find_one(
            filter={
                'key': key,
            },
        )
        if document:
            return document['value']
        else:
            return None

    def key_delete(
        self,
        key: str,
    ) -> bool:
        delete_one_result = self.connection.sergeant.keys.delete_one(
            filter={
                'key': key,
            },
        )

        return delete_one_result.deleted_count > 0

    def queue_pop(
        self,
        queue_name: str,
    ) -> typing.Optional[bytes]:
        document = self.connection.sergeant.task_queue.find_one_and_delete(
            filter={
                'queue_name': queue_name,
            },
            projection={
                'value': 1,
            },
            sort=[
                (
                    'priority',
                    pymongo.ASCENDING,
                ),
            ],
        )
        if document:
            return document['value']
        else:
            return None

    def queue_pop_bulk(
        self,
        queue_name: str,
        number_of_items: int,
    ) -> typing.List[bytes]:
        documents = []
        for i in range(number_of_items):
            document = self.connection.sergeant.task_queue.find_one_and_delete(
                filter={
                    'queue_name': queue_name,
                },
                projection={
                    'value': 1,
                },
                sort=[
                    (
                        'priority',
                        pymongo.ASCENDING,
                    ),
                ],
            )
            if document:
                documents.append(document['value'])
            else:
                break

        return documents

    def queue_push(
        self,
        queue_name: str,
        item: bytes,
        priority: str = 'NORMAL',
    ) -> bool:
        if priority == 'HIGH':
            priority_value = 0
        elif priority == 'NORMAL':
            priority_value = 1

        insert_one_result = self.connection.sergeant.task_queue.insert_one(
            document={
                'queue_name': queue_name,
                'priority': priority_value,
                'value': item,
            }
        )

        return insert_one_result.acknowledged

    def queue_push_bulk(
        self,
        queue_name: str,
        items: typing.Iterable[bytes],
        priority: str = 'NORMAL',
    ) -> bool:
        if priority == 'HIGH':
            priority_value = 0
        elif priority == 'NORMAL':
            priority_value = 1

        insert_many_result = self.connection.sergeant.task_queue.insert_many(
            documents=[
                {
                    'queue_name': queue_name,
                    'priority': priority_value,
                    'value': item,
                }
                for item in items
            ]
        )

        return insert_many_result.acknowledged

    def queue_length(
        self,
        queue_name: str,
    ) -> int:
        queue_length = self.connection.sergeant.task_queue.count_documents(
            filter={
                'queue_name': queue_name,
            },
        )

        return queue_length

    def queue_delete(
        self,
        queue_name: str,
    ) -> bool:
        result = self.connection.sergeant.task_queue.delete_many(
            filter={
                'queue_name': queue_name,
            },
        )

        return result.deleted_count > 0



import unittest
import unittest.mock
import time
import unittest
import unittest.mock

import sergeant.executor
import sergeant.config
import sergeant.task_queue


class SerialTestCase(
    unittest.TestCase,
):
    def setUp(
        self,
    ):
        self.worker = unittest.mock.MagicMock()
        self.worker.config = sergeant.config.WorkerConfig(
            name='test_worker',
            connector=sergeant.config.Connector(
                type='',
                params={},
            ),
        )
        self.worker.work = unittest.mock.MagicMock(
            return_value=True,
        )
        self.worker._on_success = unittest.mock.MagicMock()
        self.worker._on_timeout = unittest.mock.MagicMock()
        self.worker._on_failure = unittest.mock.MagicMock()
        self.worker._on_retry = unittest.mock.MagicMock()
        self.worker._on_max_retries = unittest.mock.MagicMock()
        self.worker._requeue = unittest.mock.MagicMock()

    def test_success(
        self,
    ):
        serial_executor = sergeant.executor.serial.SerialExecutor(
            worker=self.worker,
        )

        task = sergeant.task_queue.TaskQueue.craft_task(
            task_name='test_worker',
            kwargs={},
        )
        serial_executor.execute_tasks(
            tasks=[task],
        )
        self.worker.work.assert_called_once_with(
            task=task,
        )
        self.worker._on_success.assert_called_once_with(
            task=task,
            returned_value=True,
        )
        self.worker._on_failure.assert_not_called()
        self.worker._on_timeout.assert_not_called()
        self.worker._on_retry.assert_not_called()
        self.worker._on_max_retries.assert_not_called()
        self.worker._on_requeue.assert_not_called()
        self.assertIsNone(
            obj=serial_executor.killer,
        )

    def test_failure(
        self,
    ):
        def raise_exception_work_method(
            task,
        ):
            raise Exception('some exception')

        self.worker.work = unittest.mock.MagicMock(
            side_effect=lambda task: raise_exception_work_method(task),
        )

        serial_executor = sergeant.executor.serial.SerialExecutor(
            worker=self.worker,
        )

        task = sergeant.task_queue.TaskQueue.craft_task(
            task_name='test_worker',
            kwargs={},
        )
        serial_executor.execute_tasks(
            tasks=[task],
        )
        self.worker.work.assert_called_once_with(
            task=task,
        )
        self.worker._on_failure.assert_called_once()
        self.assertEqual(
            first=self.worker._on_failure.call_args[1]['task'],
            second=task,
        )
        self.assertIsInstance(
            obj=self.worker._on_failure.call_args[1]['exception'],
            cls=Exception,
        )
        self.assertEqual(
            first=self.worker._on_failure.call_args[1]['exception'].args,
            second=(
                'some exception',
            ),
        )
        self.worker._on_sucess.assert_not_called()
        self.worker._on_timeout.assert_not_called()
        self.worker._on_retry.assert_not_called()
        self.worker._on_max_retries.assert_not_called()
        self.worker._on_requeue.assert_not_called()
        self.assertIsNone(
            obj=serial_executor.killer,
        )

    def test_soft_timeout(
        self,
    ):
        def timeout_work_method(
            task,
        ):
            while True:
                time.sleep(0.1)

        self.worker.work = unittest.mock.MagicMock(
            side_effect=lambda task: timeout_work_method(task),
        )
        self.worker.config.timeouts.soft_timeout = 1.0

        serial_executor = sergeant.executor.serial.SerialExecutor(
            worker=self.worker,
        )

        task = sergeant.task_queue.TaskQueue.craft_task(
            task_name='test_worker',
            kwargs={},
        )
        serial_executor.execute_tasks(
            tasks=[task],
        )
        self.worker.work.assert_called_once_with(
            task=task,
        )
        self.worker._on_timeout.assert_called_once_with(
            task=task,
        )
        self.worker._on_sucess.assert_not_called()
        self.worker._on_failure.assert_not_called()
        self.worker._on_retry.assert_not_called()
        self.worker._on_max_retries.assert_not_called()
        self.worker._on_requeue.assert_not_called()
        self.assertIsNotNone(
            obj=serial_executor.killer,
        )

    def test_hard_timeout(
        self,
    ):
        def timeout_work_method(
            task,
        ):
            while True:
                time.sleep(0.1)

        self.worker.work = unittest.mock.MagicMock(
            side_effect=lambda task: timeout_work_method(task),
        )
        self.worker.config.timeouts.hard_timeout = 1.0

        serial_executor = sergeant.executor.serial.SerialExecutor(
            worker=self.worker,
        )

        task = sergeant.task_queue.TaskQueue.craft_task(
            task_name='test_worker',
            kwargs={},
        )
        serial_executor.execute_tasks(
            tasks=[task],
        )
        self.worker.work.assert_called_once_with(
            task=task,
        )
        self.worker._on_timeout.assert_called_once_with(
            task=task,
        )
        self.worker._on_sucess.assert_not_called()
        self.worker._on_failure.assert_not_called()
        self.worker._on_retry.assert_not_called()
        self.worker._on_max_retries.assert_not_called()
        self.worker._on_requeue.assert_not_called()
        self.assertIsNotNone(
            obj=serial_executor.killer,
        )

    def test_multiple_timeout(
        self,
    ):
        def timeout_work_method(
            task,
        ):
            while True:
                time.sleep(0.1)

        self.worker.work = unittest.mock.MagicMock(
            side_effect=lambda task: timeout_work_method(task),
        )
        self.worker.config.timeouts.soft_timeout = 1.0

        serial_executor = sergeant.executor.serial.SerialExecutor(
            worker=self.worker,
        )

        task = sergeant.task_queue.TaskQueue.craft_task(
            task_name='test_worker',
            kwargs={},
        )
        serial_executor.execute_tasks(
            tasks=[task] * 2,
        )
        self.assertEqual(
            first=self.worker.work.call_count,
            second=2,
        )
        self.assertEqual(
            first=self.worker._on_timeout.call_count,
            second=2,
        )
        self.worker._on_sucess.assert_not_called()
        self.worker._on_failure.assert_not_called()
        self.worker._on_retry.assert_not_called()
        self.worker._on_max_retries.assert_not_called()
        self.worker._on_requeue.assert_not_called()
        self.assertIsNotNone(
            obj=serial_executor.killer,
        )

    def test_on_retry(
        self,
    ):
        def retry_work_method(
            task,
        ):
            raise sergeant.worker.WorkerRetry()

        self.worker.work = unittest.mock.MagicMock(
            side_effect=lambda task: retry_work_method(task),
        )

        serial_executor = sergeant.executor.serial.SerialExecutor(
            worker=self.worker,
        )

        task = sergeant.task_queue.TaskQueue.craft_task(
            task_name='test_worker',
            kwargs={},
        )
        serial_executor.execute_tasks(
            tasks=[task],
        )
        self.worker.work.assert_called_once_with(
            task=task,
        )
        self.worker._on_retry.assert_called_once_with(
            task=task,
        )
        self.worker._on_sucess.assert_not_called()
        self.worker._on_failure.assert_not_called()
        self.worker._on_timeout.assert_not_called()
        self.worker._on_max_retries.assert_not_called()
        self.worker._on_requeue.assert_not_called()
        self.assertIsNone(
            obj=serial_executor.killer,
        )

    def test_on_max_retries(
        self,
    ):
        def max_retries_work_method(
            task,
        ):
            raise sergeant.worker.WorkerMaxRetries()

        self.worker.work = unittest.mock.MagicMock(
            side_effect=lambda task: max_retries_work_method(task),
        )

        serial_executor = sergeant.executor.serial.SerialExecutor(
            worker=self.worker,
        )

        task = sergeant.task_queue.TaskQueue.craft_task(
            task_name='test_worker',
            kwargs={},
        )
        serial_executor.execute_tasks(
            tasks=[task],
        )
        self.worker.work.assert_called_once_with(
            task=task,
        )
        self.worker._on_max_retries.assert_called_once_with(
            task=task,
        )
        self.worker._on_sucess.assert_not_called()
        self.worker._on_failure.assert_not_called()
        self.worker._on_timeout.assert_not_called()
        self.worker._on_retry.assert_not_called()
        self.worker._on_requeue.assert_not_called()
        self.assertIsNone(
            obj=serial_executor.killer,
        )

    def test_on_requeue(
        self,
    ):
        def requeue_work_method(
            task,
        ):
            raise sergeant.worker.WorkerRequeue()

        self.worker.work = unittest.mock.MagicMock(
            side_effect=lambda task: requeue_work_method(task),
        )

        serial_executor = sergeant.executor.serial.SerialExecutor(
            worker=self.worker,
        )

        task = sergeant.task_queue.TaskQueue.craft_task(
            task_name='test_worker',
            kwargs={},
        )
        serial_executor.execute_tasks(
            tasks=[task],
        )
        self.worker.work.assert_called_once_with(
            task=task,
        )
        self.worker._on_requeue.assert_called_once_with(
            task=task,
        )
        self.worker._on_sucess.assert_not_called()
        self.worker._on_failure.assert_not_called()
        self.worker._on_timeout.assert_not_called()
        self.worker._on_retry.assert_not_called()
        self.worker._on_max_retries.assert_not_called()
        self.assertIsNone(
            obj=serial_executor.killer,
        )



import typing
import time
import datetime


class TaskQueue:
    def __init__(
        self,
        connector,
        encoder,
    ) -> None:
        self.connector = connector
        self.encoder = encoder

    def purge_tasks(
        self,
        task_name: str,
    ) -> bool:
        return self.connector.queue_delete(
            queue_name=task_name,
        )

    def number_of_enqueued_tasks(
        self,
        task_name: str,
    ) -> int:
        number_of_enqueued_tasks = self.connector.queue_length(
            queue_name=task_name,
        )

        return number_of_enqueued_tasks

    @staticmethod
    def craft_task(
        task_name: str,
        kwargs: typing.Optional[typing.Dict[str, typing.Any]] = None,
    ) -> typing.Dict[str, typing.Any]:
        if kwargs is None:
            kwargs = {}

        task = {
            'name': task_name,
            'date': datetime.datetime.utcnow().timestamp(),
            'kwargs': kwargs,
            'run_count': 0,
        }

        return task

    def wait_queue_empty(
        self,
        task_name: str,
        timeout: float = 0,
        sample_interval: float = 1.0,
    ) -> bool:
        remaining_time = timeout

        not_empty = True
        while not_empty:
            if timeout and remaining_time <= 0:
                return False

            not_empty = self.number_of_enqueued_tasks(
                task_name=task_name,
            ) != 0

            time.sleep(sample_interval)
            remaining_time -= sample_interval

        return True

    def apply_async_one(
        self,
        task: typing.Dict[str, typing.Any],
        priority: str = 'NORMAL',
    ) -> bool:
        encoded_item = self.encoder.encode(
            data=task,
        )

        pushed = self.connector.queue_push(
            queue_name=task['name'],
            item=encoded_item,
            priority=priority,
        )

        return pushed

    def apply_async_many(
        self,
        task_name: str,
        tasks: typing.Iterable[typing.Dict[str, typing.Any]],
        priority: str = 'NORMAL',
    ) -> bool:
        encoded_tasks = []
        for task in tasks:
            encoded_task = self.encoder.encode(
                data=task,
            )

            encoded_tasks.append(encoded_task)

        if encoded_tasks:
            self.connector.queue_push_bulk(
                queue_name=task_name,
                items=encoded_tasks,
                priority=priority,
            )

        return True

    def get_tasks(
        self,
        task_name: str,
        number_of_tasks: int,
    ) -> typing.List[typing.Dict[str, typing.Any]]:
        if number_of_tasks == 1:
            task = self.connector.queue_pop(
                queue_name=task_name,
            )
            if not task:
                return []
            else:
                tasks = [task]
        else:
            tasks = self.connector.queue_pop_bulk(
                queue_name=task_name,
                number_of_items=number_of_tasks,
            )

        decoded_tasks = [
            self.encoder.decode(
                data=task,
            )
            for task in tasks
        ]

        return decoded_tasks

    def retry(
        self,
        task: typing.Dict[str, typing.Any],
        priority: str = 'NORMAL',
    ) -> bool:
        task['run_count'] += 1

        return self.apply_async_one(
            task=task,
            priority=priority,
        )

    def requeue(
        self,
        task: typing.Dict[str, typing.Any],
        priority='NORMAL',
    ) -> bool:
        return self.apply_async_one(
            task=task,
            priority=priority,
        )

    def delete_key(
        self,
        name: str,
    ) -> bool:
        key_was_deleted = self.connector.key_delete(
            key=name.encode(),
        )

        return key_was_deleted

    def get_key(
        self,
        name: str,
    ) -> typing.Any:
        value = self.connector.key_get(
            key=name.encode(),
        )
        if not value:
            return value

        decoded_value = self.encoder.decode(
            data=value,
        )

        return decoded_value

    def set_key(
        self,
        name: str,
        value: typing.Any,
    ) -> bool:
        encoded_value = self.encoder.encode(
            data=value,
        )

        key_was_set = self.connector.key_set(
            key=name.encode(),
            value=encoded_value,
        )

        return key_was_set



import typing
import orjson


class Serializer:
    name = 'json'

    @staticmethod
    def serialize(
        data: typing.Any,
    ) -> bytes:
        return orjson.dumps(data)

    @staticmethod
    def unserialize(
        data: bytes,
    ) -> typing.Any:
        return orjson.loads(data)



import threading
import time
import ctypes


class ThreadKiller(
    threading.Thread,
):
    def __init__(
        self,
        thread_id: int,
        timeout: float,
        exception: Exception,
    ) -> None:
        super().__init__()

        self.timeout = timeout
        self.exception = exception
        self.thread_id = thread_id

        self.time_elapsed = 0.0
        self.enabled = threading.Event()
        self.enabled.set()
        self.running = threading.Event()
        self.running.set()

    def kill(
        self,
    ) -> None:
        self.enabled.clear()

    def suspend(
        self,
    ) -> None:
        self.running.clear()

    def resume(
        self,
    ) -> None:
        self.running.set()

    def reset(
        self,
    ) -> None:
        self.time_elapsed = 0

    def run(
        self,
    ) -> None:
        while self.enabled.is_set():
            if self.running.is_set():
                if self.time_elapsed >= self.timeout:
                    ctypes.pythonapi.PyThreadState_SetAsyncExc(
                        ctypes.c_long(self.thread_id),
                        ctypes.py_object(self.exception),
                    )

                    return

                self.time_elapsed += 0.1

            time.sleep(0.1)



import time
import unittest
import threading

import sergeant.devices.thread_killer


class ThreadKillerTestCase(
    unittest.TestCase,
):
    def setUp(
        self,
    ):
        self.raised_exception = None

    def tearDown(
        self,
    ):
        pass

    def thread_function(
        self,
        interval,
    ):
        try:
            while True:
                time.sleep(0.05)
        except Exception as exception:
            self.raised_exception = exception

    def test_start(
        self,
    ):
        thread = threading.Thread(
            target=self.thread_function,
            kwargs={
                'interval': 2,
            },
        )
        thread.start()

        killer = sergeant.devices.thread_killer.ThreadKiller(
            thread_id=thread.ident,
            timeout=0.5,
            exception=ExceptionTest,
        )
        killer.start()
        self.assertTrue(
            expr=thread.is_alive(),
        )
        time.sleep(0.6)
        self.assertFalse(
            expr=thread.is_alive(),
        )
        self.assertIsInstance(
            obj=self.raised_exception,
            cls=ExceptionTest,
        )

    def test_stop(
        self,
    ):
        thread = threading.Thread(
            target=self.thread_function,
            kwargs={
                'interval': 5,
            },
        )
        thread.start()

        killer = sergeant.devices.thread_killer.ThreadKiller(
            thread_id=thread.ident,
            timeout=1.0,
            exception=ExceptionTest,
        )
        killer.start()
        self.assertTrue(
            expr=thread.is_alive(),
        )
        time.sleep(0.5)
        killer.suspend()
        time.sleep(1.0)
        self.assertTrue(
            expr=thread.is_alive(),
        )
        killer.resume()
        time.sleep(1.0)
        self.assertFalse(
            expr=thread.is_alive(),
        )

        self.assertIsInstance(
            obj=self.raised_exception,
            cls=ExceptionTest,
        )

    def test_reset(
        self,
    ):
        thread = threading.Thread(
            target=self.thread_function,
            kwargs={
                'interval': 5,
            },
        )
        thread.start()

        killer = sergeant.devices.thread_killer.ThreadKiller(
            thread_id=thread.ident,
            timeout=1.0,
            exception=ExceptionTest,
        )
        killer.start()

        self.assertTrue(
            expr=thread.is_alive(),
        )
        time.sleep(0.9)
        killer.reset()
        time.sleep(0.9)
        self.assertTrue(
            expr=thread.is_alive(),
        )
        time.sleep(0.4)
        self.assertFalse(
            expr=thread.is_alive(),
        )
        self.assertIsInstance(
            obj=self.raised_exception,
            cls=ExceptionTest,
        )


class ExceptionTest(
    Exception,
):
    pass



import typing
import sys
import os
import multiprocessing
import subprocess
import shlex
import multiprocessing.context
import argparse
import logging
import psutil
import time


class SupervisedWorker:
    def __init__(
        self,
        worker_module_name: str,
        worker_class_name: str,
    ):
        self.logger = logging.getLogger(
            name='SupervisedWorker',
        )

        pipe = multiprocessing.Pipe()

        self.parent_pipe = pipe[0]
        self.child_pipe = pipe[1]

        self.process = subprocess.Popen(
            args=shlex.split(
                s=f'python3 -m sergeant.slave --child-pipe={self.child_pipe.fileno()} --worker-module={worker_module_name} --worker-class={worker_class_name}',
            ),
            pass_fds=(
                self.child_pipe.fileno(),
            ),
        )

        self.psutil_obj = psutil.Process(
            pid=self.process.pid,
        )

    def get_rss_memory(
        self,
    ) -> int:
        try:
            return self.psutil_obj.memory_info().rss
        except psutil.NoSuchProcess:
            return 0

    def get_summary(
        self,
    ) -> typing.Optional[typing.Dict[str, typing.Any]]:
        try:
            if self.parent_pipe.poll():
                return self.parent_pipe.recv()
            else:
                return None
        except Exception as exception:
            self.logger.error(f'could not receive supervised_worker\'s summary: {exception}')

            return None

    def kill(
        self,
    ) -> None:
        try:
            self.child_pipe.close()
            self.parent_pipe.close()
        except Exception:
            pass

        try:
            self.process.terminate()
        except Exception:
            pass

        try:
            self.process.wait()
        except (
            ChildProcessError,
            OSError,
        ):
            pass

    def __del__(
        self,
    ) -> None:
        self.kill()


class Supervisor:
    def __init__(
        self,
        worker_module_name: str,
        worker_class_name: str,
        concurrent_workers: int,
        max_worker_memory_usage: typing.Optional[int],
    ):
        self.worker_module_name = worker_module_name
        self.worker_class_name = worker_class_name
        self.concurrent_workers = concurrent_workers
        self.max_worker_memory_usage = max_worker_memory_usage

        self.logger = logging.getLogger(
            name='Supervisor',
        )

        formatter = logging.Formatter(
            fmt='%(asctime)s %(name)-12s %(process)d %(levelname)-8s %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S',
        )
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(formatter)
        self.logger.addHandler(stream_handler)
        self.logger.setLevel(logging.INFO)

        self.current_workers = []

    def start(
        self,
    ) -> None:
        for i in range(self.concurrent_workers):
            worker = SupervisedWorker(
                worker_module_name=self.worker_module_name,
                worker_class_name=self.worker_class_name,
            )
            self.logger.info(f'spawned a new worker at pid: {worker.process.pid}')

            self.current_workers.append(worker)

        self.supervise_loop()

    def supervise_loop(
        self,
    ) -> None:
        try:
            while True:
                current_workers = self.current_workers.copy()
                for worker in current_workers:
                    self.supervise_worker(
                        worker=worker,
                    )

                time.sleep(0.5)
        except KeyboardInterrupt:
            pass
        except Exception as exception:
            self.logger.error(f'the supervisor encountered an exception: {exception}')
            self.logger.error(f'exiting...')
        finally:
            for worker in self.current_workers:
                self.logger.info(f'killing a worker: {worker.process.pid}')
                worker.kill()
            sys.exit(0)

    def supervise_worker(
        self,
        worker: SupervisedWorker,
    ) -> None:
        if worker.process.poll() is not None:
            worker_summary = worker.get_summary()

            self.logger.info(f'a worker has exited with error code: {worker.process.returncode}')

            try:
                os.waitpid(worker.process.pid, 0)
            except ChildProcessError:
                pass

            if worker.process.returncode == 0:
                if worker_summary:
                    self.logger.debug(f'worker summary: {worker_summary}')
                else:
                    self.logger.debug(f'worker summary is unavailable')

            if worker.process.returncode == 1:
                self.logger.error(f'worker execution has failed')

                if worker_summary:
                    self.logger.error(f'exception: {worker_summary["exception"]}')
                    self.logger.error(f'traceback: {worker_summary["traceback"]}')
                else:
                    self.logger.error(f'exception and tracback are unavailable')

            if worker.process.returncode == 2:
                self.logger.error(f'could not load worker module: {self.worker_module_name}')

                sys.exit(1)

            if worker.process.returncode == 3:
                self.logger.error(f'could not find worker class: {self.worker_module_name}.{self.worker_class_name}')

                sys.exit(1)

            self.respawn_a_worker(
                worker=worker,
            )

        if self.max_worker_memory_usage:
            if worker.get_rss_memory() > self.max_worker_memory_usage:
                self.logger.info(f'worker exceeded the maximum memory limit: pid: {worker.process.pid}, rss: {worker.get_rss_memory()}')
                self.respawn_a_worker(
                    worker=worker,
                )

    def respawn_a_worker(
        self,
        worker: SupervisedWorker,
    ) -> None:
        worker.kill()
        self.current_workers.remove(worker)
        worker = SupervisedWorker(
            worker_module_name=self.worker_module_name,
            worker_class_name=self.worker_class_name,
        )
        self.logger.info(f'spawned a new worker at pid: {worker.process.pid}')
        self.current_workers.append(worker)


def main() -> None:
    parser = argparse.ArgumentParser(
        description='Sergeant Supervisor',
    )
    parser.add_argument(
        '--concurrent-workers',
        help='Number of subprocesses to open',
        type=int,
        required=True,
        dest='concurrent_workers',
    )
    parser.add_argument(
        '--worker-class',
        help='Class name of the worker to spawn',
        type=str,
        required=True,
        dest='worker_class',
    )
    parser.add_argument(
        '--worker-module',
        help='Module of the worker class',
        type=str,
        required=True,
        dest='worker_module',
    )
    parser.add_argument(
        '--max-worker-memory-usage',
        help='Maximum RSS memory usage in bytes of an individual worker. When a worker reaches this value, the supevisor would kill it and respawn another one in place.',
        type=int,
        required=False,
        dest='max_worker_memory_usage',
    )
    args = parser.parse_args()

    supervisor = Supervisor(
        worker_module_name=args.worker_module,
        worker_class_name=args.worker_class,
        concurrent_workers=args.concurrent_workers,
        max_worker_memory_usage=args.max_worker_memory_usage,
    )
    supervisor.start()


if __name__ == '__main__':
    main()



import zlib


class Compressor:
    name = 'zlib'

    @staticmethod
    def compress(
        data: bytes,
    ) -> bytes:
        compressed_object = zlib.compress(data)

        return compressed_object

    @staticmethod
    def decompress(
        data: bytes,
    ) -> bytes:
        decompressed_object = zlib.decompress(data)

        return decompressed_object



import typing
import threading
import concurrent.futures

from .. import worker
from .. import devices


class ThreadedExecutor:
    def __init__(
        self,
        worker: worker.Worker,
        number_of_threads: int,
    ) -> None:
        self.worker = worker
        self.number_of_threads = number_of_threads

        has_soft_timeout = self.worker.config.timeouts.soft_timeout > 0
        self.should_use_a_killer = has_soft_timeout

    def execute_tasks(
        self,
        tasks: typing.Iterable[typing.Dict[str, typing.Any]],
    ) -> None:
        future_to_task = {}
        self.thread_killers = {}

        with concurrent.futures.ThreadPoolExecutor(
            max_workers=self.number_of_threads,
        ) as executor:
            for task in tasks:
                future = executor.submit(self.execute_task, task)
                future_to_task[future] = task

                if len(future_to_task) == self.number_of_threads:
                    finished, pending = concurrent.futures.wait(
                        fs=future_to_task,
                        timeout=None,
                        return_when=concurrent.futures.FIRST_COMPLETED,
                    )
                    for finished_future in finished:
                        del future_to_task[finished_future]

            for future in concurrent.futures.as_completed(future_to_task):
                pass

        for thread_killer in self.thread_killers.values():
            thread_killer.kill()
        self.thread_killers.clear()

    def execute_task(
        self,
        task: typing.Dict[str, typing.Any],
    ) -> None:
        try:
            self.pre_work(
                task=task,
            )

            returned_value = self.worker.work(
                task=task,
            )

            self.post_work(
                task=task,
                success=True,
            )

            self.worker._on_success(
                task=task,
                returned_value=returned_value,
            )
        except Exception as exception:
            self.post_work(
                task=task,
                success=False,
                exception=exception,
            )

            if isinstance(exception, worker.WorkerTimedout):
                self.worker._on_timeout(
                    task=task,
                )
            elif isinstance(exception, worker.WorkerRetry):
                self.worker._on_retry(
                    task=task,
                )
            elif isinstance(exception, worker.WorkerMaxRetries):
                self.worker._on_max_retries(
                    task=task,
                )
            elif isinstance(exception, worker.WorkerRequeue):
                self.worker._on_requeue(
                    task=task,
                )
            else:
                self.worker._on_failure(
                    task=task,
                    exception=exception,
                )

    def pre_work(
        self,
        task: typing.Dict[str, typing.Any],
    ) -> None:
        try:
            self.worker.pre_work(
                task=task,
            )
        except Exception as exception:
            self.worker.logger.error(
                msg=f'pre_work has failed: {exception}',
                extra={
                    'task': task,
                },
            )

        if self.should_use_a_killer:
            current_thread_id = threading.get_ident()

            if current_thread_id in self.thread_killers:
                self.thread_killers[current_thread_id].reset()
                self.thread_killers[current_thread_id].resume()
            else:
                self.thread_killers[current_thread_id] = devices.thread_killer.ThreadKiller(
                    thread_id=current_thread_id,
                    timeout=self.worker.config.timeouts.soft_timeout,
                    exception=worker.WorkerSoftTimedout,
                )
                self.thread_killers[current_thread_id].start()

    def post_work(
        self,
        task: typing.Dict[str, typing.Any],
        success: bool,
        exception: typing.Optional[Exception] = None,
    ) -> None:
        if self.should_use_a_killer:
            self.thread_killers[threading.get_ident()].suspend()

        try:
            self.worker.post_work(
                task=task,
                success=success,
                exception=exception,
            )
        except Exception as exception:
            self.worker.logger.error(
                msg=f'post_work has failed: {exception}',
                extra={
                    'task': task,
                },
            )



import logging
import socket
import datetime
import orjson
import sys
import traceback


class LogstashHandler(
    logging.Handler,
):
    def __init__(
        self,
        host: str,
        port: int,
    ) -> None:
        super().__init__()
        self.hostname = socket.gethostname()
        self.ipaddress = socket.gethostbyname(self.hostname)
        self.host = host
        self.port = port
        self.address = (
            self.host,
            self.port,
        )

    def emit(
        self,
        record: logging.LogRecord,
    ) -> None:
        message = {
            '@timestamp': datetime.datetime.utcfromtimestamp(record.created).strftime('%Y-%m-%dT%H:%M:%SZ'),
            'message': record.getMessage(),
            'logging': {
                'level': record.levelname,
                'name': record.name,
            },
            'emitter': {
                'hostname': self.hostname,
                'ipaddress': self.ipaddress,
                'filename': record.filename,
                'function': record.funcName,
                'line': record.lineno,
            },
            'task': getattr(record, 'task', None),
        }

        last_exception_type, last_exception_object, last_exception_tb = sys.exc_info()
        if last_exception_object:
            extracted_stacktrace = traceback.format_exception(
                etype=last_exception_type,
                value=last_exception_object,
                tb=last_exception_tb,
            )
            message['exception'] = {
                'type': last_exception_object.__class__.__name__,
                'message': str(last_exception_object),
                'stacktrace': ''.join(extracted_stacktrace),
            }

        encoded_message = orjson.dumps(
            obj=message,
            default=repr,
        )

        try:
            socket_connection = socket.socket(
                family=socket.AF_INET,
                type=socket.SOCK_STREAM,
            )
            socket_connection.connect(self.address)
            socket_connection.sendall(encoded_message)
        except Exception:
            traceback.print_exc()
        finally:
            try:
                socket_connection.close()
            except Exception:
                pass



import multiprocessing
import os
import signal
import sys
import time
import unittest

import sergeant.devices.process_killer


class ProcessKillerTestCase(
    unittest.TestCase,
):
    def setUp(
        self,
    ):
        self.sigint_fired = False
        self.sigabrt_fired = False

        signal.signal(
            signalnum=signal.SIGABRT,
            handler=self.sigabrt_handler,
        )
        signal.signal(
            signalnum=signal.SIGINT,
            handler=self.sigint_handler,
        )

    def tearDown(
        self,
    ):
        for signal_type in [
            signal.SIGABRT,
            signal.SIGINT,
        ]:
            signal.signal(
                signalnum=signal_type,
                handler=signal.SIG_DFL,
            )

    def sigabrt_handler(
        self,
        signal_num,
        frame,
    ):
        self.sigabrt_fired = True

    def sigint_handler(
        self,
        signal_num,
        frame,
    ):
        self.sigint_fired = True

    def test_timeouts_killer(
        self,
    ):
        killer = sergeant.devices.process_killer.ProcessKiller(
            pid_to_kill=os.getpid(),
            sleep_interval=0.05,
            soft_timeout=1.0,
            hard_timeout=3.0,
            critical_timeout=5.0,
        )

        killer.start()
        self.assertFalse(
            expr=self.sigabrt_fired,
        )
        self.assertFalse(
            expr=self.sigint_fired,
        )
        time.sleep(1.2)
        self.assertFalse(
            expr=self.sigabrt_fired,
        )
        self.assertTrue(
            expr=self.sigint_fired,
        )
        time.sleep(1.2)
        self.assertFalse(
            expr=self.sigabrt_fired,
        )
        self.assertTrue(
            expr=self.sigint_fired,
        )
        time.sleep(1.2)
        self.assertTrue(
            expr=self.sigabrt_fired,
        )
        self.assertTrue(
            expr=self.sigint_fired,
        )
        killer.stop()

        self.sigint_fired = False
        self.sigabrt_fired = False
        time.sleep(1.2)
        self.assertFalse(
            expr=self.sigabrt_fired,
        )
        self.assertFalse(
            expr=self.sigint_fired,
        )
        time.sleep(1.2)
        self.assertFalse(
            expr=self.sigabrt_fired,
        )
        self.assertFalse(
            expr=self.sigint_fired,
        )

        killer.reset()
        killer.start()
        self.assertFalse(
            expr=self.sigabrt_fired,
        )
        self.assertFalse(
            expr=self.sigint_fired,
        )
        time.sleep(0.5)
        self.assertFalse(
            expr=self.sigabrt_fired,
        )
        self.assertFalse(
            expr=self.sigint_fired,
        )
        killer.reset()
        time.sleep(0.5)
        self.assertFalse(
            expr=self.sigabrt_fired,
        )
        self.assertFalse(
            expr=self.sigint_fired,
        )
        killer.reset()
        time.sleep(0.5)
        self.assertFalse(
            expr=self.sigabrt_fired,
        )
        self.assertFalse(
            expr=self.sigint_fired,
        )
        killer.reset()
        time.sleep(0.5)
        self.assertFalse(
            expr=self.sigabrt_fired,
        )
        self.assertFalse(
            expr=self.sigint_fired,
        )
        killer.reset()
        killer.stop()
        self.assertFalse(
            expr=self.sigabrt_fired,
        )
        self.assertFalse(
            expr=self.sigint_fired,
        )

    def test_sleep_case_killer(
        self,
    ):
        test_process_obj = TestProcess()
        testing_process = multiprocessing.Process(
            target=test_process_obj.sleep,
            kwargs={
                'interval': 30,
            },
        )
        testing_process.daemon = True
        testing_process.start()

        self.assertTrue(
            expr=testing_process.is_alive(),
        )

        killer = sergeant.devices.process_killer.ProcessKiller(
            pid_to_kill=testing_process.pid,
            sleep_interval=0.05,
            soft_timeout=1.0,
            hard_timeout=2.0,
            critical_timeout=5.0,
        )

        killer.start()
        self.assertTrue(
            expr=testing_process.is_alive(),
        )
        time.sleep(1.2)
        self.assertFalse(
            expr=testing_process.is_alive(),
        )
        self.assertEqual(testing_process.exitcode, 20)
        killer.stop()

    def test_no_int_case_killer(
        self,
    ):
        test_process_obj = TestProcess()
        testing_process = multiprocessing.Process(
            target=test_process_obj.no_int_sleep,
            kwargs={
                'interval': 30,
            },
        )
        testing_process.daemon = True
        testing_process.start()

        self.assertTrue(
            expr=testing_process.is_alive(),
        )

        killer = sergeant.devices.process_killer.ProcessKiller(
            pid_to_kill=testing_process.pid,
            sleep_interval=0.05,
            soft_timeout=1.0,
            hard_timeout=2.0,
            critical_timeout=5.0,
        )

        killer.start()
        self.assertTrue(
            expr=testing_process.is_alive(),
        )
        time.sleep(1.2)
        self.assertTrue(
            expr=testing_process.is_alive(),
        )
        time.sleep(1.2)
        self.assertFalse(
            expr=testing_process.is_alive(),
        )
        self.assertEqual(testing_process.exitcode, 10)
        killer.stop()

    def test_lost_case_killer(
        self,
    ):
        test_process_obj = TestProcess()
        testing_process = multiprocessing.Process(
            target=test_process_obj.lost,
            kwargs={
                'interval': 30,
            },
        )
        testing_process.daemon = True
        testing_process.start()

        self.assertTrue(
            expr=testing_process.is_alive(),
        )

        killer = sergeant.devices.process_killer.ProcessKiller(
            pid_to_kill=testing_process.pid,
            sleep_interval=0.05,
            soft_timeout=1.0,
            hard_timeout=2.0,
            critical_timeout=3.0,
        )

        killer.start()
        self.assertTrue(
            expr=testing_process.is_alive(),
        )
        time.sleep(1.2)
        self.assertTrue(
            expr=testing_process.is_alive(),
        )
        time.sleep(1.2)
        self.assertTrue(
            expr=testing_process.is_alive(),
        )
        time.sleep(1.2)
        self.assertFalse(
            expr=testing_process.is_alive(),
        )
        self.assertEqual(testing_process.exitcode, -15)
        killer.stop()


class TestProcess:
    def init(
        self,
    ):
        signal.signal(
            signalnum=signal.SIGABRT,
            handler=self.sigabrt_handler,
        )
        signal.signal(
            signalnum=signal.SIGINT,
            handler=self.sigint_handler,
        )

    def sleep(
        self,
        interval,
    ):
        self.init()
        time.sleep(interval)

    def no_int_sleep(
        self,
        interval,
    ):
        self.init()
        signal.signal(
            signalnum=signal.SIGINT,
            handler=lambda a, b: True,
        )
        time.sleep(interval)

    def lost(
        self,
        interval,
    ):
        self.init()
        signal.signal(
            signalnum=signal.SIGINT,
            handler=lambda a, b: True,
        )
        signal.signal(
            signalnum=signal.SIGABRT,
            handler=lambda a, b: True,
        )
        time.sleep(interval)

    def sigabrt_handler(
        self,
        signal_num,
        frame,
    ):
        sys.exit(10)

    def sigint_handler(
        self,
        signal_num,
        frame,
    ):
        sys.exit(20)

    def __setstate__(
        self,
        state,
    ):
        self.init()



import unittest

import sergeant.encoder.compressor


class CompressorsTestCase(
    unittest.TestCase,
):
    obj_to_compress = b''.join(chr(i).encode() for i in range(2**15))

    def test_bzip2(
        self,
    ):
        compressed_object = sergeant.encoder.compressor.bzip2.Compressor.compress(
            data=self.obj_to_compress,
        )
        uncompressed_object = sergeant.encoder.compressor.bzip2.Compressor.decompress(
            data=compressed_object,
        )

        self.assertEqual(
            first=uncompressed_object,
            second=self.obj_to_compress,
        )

    def test_gzip(
        self,
    ):
        compressed_object = sergeant.encoder.compressor.gzip.Compressor.compress(
            data=self.obj_to_compress,
        )
        uncompressed_object = sergeant.encoder.compressor.gzip.Compressor.decompress(
            data=compressed_object,
        )

        self.assertEqual(
            first=uncompressed_object,
            second=self.obj_to_compress,
        )

    def test_lzma(
        self,
    ):
        compressed_object = sergeant.encoder.compressor.lzma.Compressor.compress(
            data=self.obj_to_compress,
        )
        uncompressed_object = sergeant.encoder.compressor.lzma.Compressor.decompress(
            data=compressed_object,
        )

        self.assertEqual(
            first=uncompressed_object,
            second=self.obj_to_compress,
        )

    def test_zlib(
        self,
    ):
        compressed_object = sergeant.encoder.compressor.zlib.Compressor.compress(
            data=self.obj_to_compress,
        )
        uncompressed_object = sergeant.encoder.compressor.zlib.Compressor.decompress(
            data=compressed_object,
        )

        self.assertEqual(
            first=uncompressed_object,
            second=self.obj_to_compress,
        )



from . import bzip2
from . import gzip
from . import lzma
from . import zlib


__compressors__ = {
    bzip2.Compressor.name: bzip2.Compressor,
    gzip.Compressor.name: gzip.Compressor,
    lzma.Compressor.name: lzma.Compressor,
    zlib.Compressor.name: zlib.Compressor,
}



import typing
import os
import signal
import asyncio
import subprocess
import shlex
import sys
import psutil
import argparse
import socket
import logging


class ProcessKillerServer(
    asyncio.DatagramProtocol,
):
    def __init__(
        self,
        async_loop: asyncio.AbstractEventLoop,
        pid_to_kill: int,
        sleep_interval: float,
        soft_timeout: float,
        hard_timeout: float,
        critical_timeout: float,
    ) -> None:
        super().__init__()

        self.logger = logging.getLogger(
            name='Killer',
        )
        self.logger.propagate = False
        self.logger.setLevel(
            level=logging.ERROR,
        )
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(
            fmt=logging.Formatter(
                fmt='%(asctime)s %(name)-12s %(process)d %(levelname)-8s %(message)s',
                datefmt='%Y-%m-%d %H:%M:%S',
            ),
        )
        self.logger.addHandler(
            hdlr=stream_handler,
        )

        self.pid_to_kill = pid_to_kill
        self.async_loop = async_loop

        self.sleep_interval = sleep_interval
        self.time_elapsed = 0.0

        self.soft_timeout = soft_timeout
        self.hard_timeout = hard_timeout
        self.critical_timeout = critical_timeout

        self.soft_timeout_raised = False
        self.hard_timeout_raised = False
        self.critical_timeout_raised = False

        self.soft_timeout_signal = signal.SIGINT
        self.hard_timeout_signal = signal.SIGABRT
        self.critical_timeout_signal = signal.SIGTERM

        self.running = False

        self.initialized = False

    def connection_made(
        self,
        transport: asyncio.BaseTransport,
    ) -> None:
        udp_port = transport._sock.getsockname()[1]
        sys.stdout.write(str(udp_port) + '\n')
        sys.stdout.flush()

    def datagram_received(
        self,
        data: bytes,
        addr: typing.Tuple[str, int],
    ) -> None:
        if data == b'start':
            self.logger.info(
                msg='start request was received',
            )
            if not self.initialized:
                self.async_loop.create_task(self.kill_loop())
                self.initialized = True

            self.killer_start()
        elif data == b'stop':
            self.logger.info(
                msg='stop request was received',
            )
            self.killer_stop()
        elif data == b'reset':
            self.logger.info(
                msg='reset request was received',
            )
            self.killer_reset()
        elif data == b'stop_and_reset':
            self.logger.info(
                msg='stop_and_reset request was received',
            )
            self.killer_stop()
            self.killer_reset()

    async def kill_loop(
        self,
    ) -> None:
        self.logger.info(
            msg='kill loop started',
        )

        process_to_kill = psutil.Process(
            pid=self.pid_to_kill,
        )
        while process_to_kill.is_running():
            try:
                process_to_kill_status = process_to_kill.status()
                if process_to_kill_status in [
                    psutil.STATUS_DEAD,
                    psutil.STATUS_ZOMBIE,
                ]:
                    self.logger.error(
                        msg=f'kill_loop process status not running: {process_to_kill_status}',
                    )

                    try:
                        process_to_kill.kill()
                        process_to_kill.wait(0)
                    except psutil.TimeoutExpired:
                        pass

                    break

                if self.running:
                    if not self.soft_timeout_raised and self.soft_timeout != 0 and self.time_elapsed >= self.soft_timeout:
                        self.logger.info(
                            msg='raising soft timeout',
                        )
                        self.soft_timeout_raised = True
                        self.kill_process(
                            pid=self.pid_to_kill,
                            signal=self.soft_timeout_signal,
                        )

                    if not self.hard_timeout_raised and self.hard_timeout != 0 and self.time_elapsed >= self.hard_timeout:
                        self.logger.info(
                            msg='raising hard timeout',
                        )
                        self.hard_timeout_raised = True
                        self.kill_process(
                            pid=self.pid_to_kill,
                            signal=self.hard_timeout_signal,
                        )

                    if not self.critical_timeout_raised and self.critical_timeout != 0 and self.time_elapsed >= self.critical_timeout:
                        self.logger.info(
                            msg='raising critical timeout',
                        )
                        self.critical_timeout_raised = True
                        self.kill_process(
                            pid=self.pid_to_kill,
                            signal=self.critical_timeout_signal,
                        )

                    self.time_elapsed += self.sleep_interval

                await asyncio.sleep(
                    delay=self.sleep_interval,
                    loop=self.async_loop,
                )
            except psutil.NoSuchProcess:
                self.logger.info(
                    msg='pid does not exist anymore, exiting',
                )
            except Exception as exception:
                self.logger.error(
                    msg=f'kill_loop exception occured: {exception}',
                )

    def kill_process(
        self,
        pid: int,
        signal: int,
    ) -> None:
        try:
            self.logger.info(
                msg='kill_process request',
            )

            os.kill(pid, signal)
        except Exception as exception:
            self.logger.error(
                msg=str(exception),
            )

    def killer_start(
        self,
    ) -> None:
        self.running = True

        self.logger.info(
            msg='started',
        )

    def killer_stop(
        self,
    ) -> None:
        self.running = False

        self.logger.info(
            msg='stopped',
        )

    def killer_reset(
        self,
    ) -> None:
        self.time_elapsed = 0.0
        self.soft_timeout_raised = False
        self.hard_timeout_raised = False
        self.critical_timeout_raised = False

        self.logger.info(
            msg='reset',
        )


class ProcessKiller:
    def __init__(
        self,
        pid_to_kill: int,
        sleep_interval: float,
        soft_timeout: float,
        hard_timeout: float,
        critical_timeout: float,
    ) -> None:
        self.killer_process = subprocess.Popen(
            args=shlex.split(
                s=f'python3 -m {__name__} --pid-to-kill {pid_to_kill} --sleep-interval {sleep_interval} --soft-timeout {soft_timeout} --hard-timeout {hard_timeout} --critical-timeout {critical_timeout}',
            ),
            cwd=os.path.join(
                os.path.dirname(
                    p=os.path.realpath(
                        filename=__file__,
                    ),
                ),
                '../../'
            ),
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
            bufsize=1,
            universal_newlines=True,
        )

        self.port = int(self.killer_process.stdout.readline().strip())
        self.killer_process.stdout.close()

        self.killer_socket = socket.socket(
            family=socket.AF_INET,
            type=socket.SOCK_DGRAM,
        )
        self.address = (
            '127.0.0.1',
            self.port,
        )
        self.killer_socket.connect(self.address)

    def start(
        self,
    ) -> None:
        self.killer_socket.send(b'start')

    def stop(
        self,
    ) -> None:
        self.killer_socket.send(b'stop')

    def reset(
        self,
    ) -> None:
        self.killer_socket.send(b'reset')

    def stop_and_reset(
        self,
    ) -> None:
        self.killer_socket.send(b'stop_and_reset')

    def kill(
        self,
    ) -> None:
        try:
            self.killer_socket.close()
        except Exception:
            pass

        try:
            self.killer_process.terminate()
        except Exception:
            pass

        try:
            self.killer_process.wait()
        except (
            ChildProcessError,
            OSError,
        ):
            pass

    def __del__(
        self,
    ) -> None:
        try:
            self.kill()
        except Exception:
            pass


def main():
    parser = argparse.ArgumentParser(
        description='Process Killer',
    )
    parser.add_argument(
        '--pid-to-kill',
        help='pid to kill on timeouts',
        type=int,
        required=True,
        dest='pid_to_kill',
    )
    parser.add_argument(
        '--sleep-interval',
        help='time to sleep between checks',
        type=float,
        required=True,
        dest='sleep_interval',
    )
    parser.add_argument(
        '--soft-timeout',
        help='soft timeout',
        type=float,
        required=True,
        dest='soft_timeout',
    )
    parser.add_argument(
        '--hard-timeout',
        help='hard timeout',
        type=float,
        required=True,
        dest='hard_timeout',
    )
    parser.add_argument(
        '--critical-timeout',
        help='critical timeout',
        type=float,
        required=True,
        dest='critical_timeout',
    )

    args = parser.parse_args()

    async_loop = asyncio.new_event_loop()

    killer_udp_server_endpoint = async_loop.create_datagram_endpoint(
        protocol_factory=lambda: ProcessKillerServer(
            async_loop=async_loop,
            pid_to_kill=args.pid_to_kill,
            sleep_interval=args.sleep_interval,
            soft_timeout=args.soft_timeout,
            hard_timeout=args.hard_timeout,
            critical_timeout=args.critical_timeout,
        ),
        local_addr=(
            '127.0.0.1',
            0,
        ),
    )

    async_loop.create_task(killer_udp_server_endpoint)

    try:
        async_loop.run_forever()
    except Exception:
        async_loop.close()


if __name__ == '__main__':
    main()



from . import mongo
from . import redis
from . import redis_cluster


__connectors__ = {
    mongo.Connector.name: mongo.Connector,
    redis_cluster.Connector.name: redis_cluster.Connector,
    redis.Connector.name: redis.Connector,
}



import typing
import pickle


class Serializer:
    name = 'pickle'

    @staticmethod
    def serialize(
        data: typing.Any,
    ) -> bytes:
        return pickle.dumps(
            obj=data,
            protocol=pickle.HIGHEST_PROTOCOL,
        )

    @staticmethod
    def unserialize(
        data: bytes,
    ) -> typing.Any:
        return pickle.loads(
            data=data,
            encoding='utf-8',
        )



import unittest
import unittest.mock

import sergeant.worker


class WorkerTestCase(
    unittest.TestCase,
):
    def test_init_logger(
        self,
    ):
        worker = sergeant.worker.Worker()
        worker.config = sergeant.config.WorkerConfig(
            name='some_worker',
            connector=sergeant.config.Connector(
                type='',
                params={},
            ),
        )

        worker.init_logger()
        self.assertIsNotNone(
            obj=worker.logger,
        )

        worker.config = sergeant.config.WorkerConfig(
            name='some_worker',
            connector=sergeant.config.Connector(
                type='',
                params={},
            ),
            logging=sergeant.config.Logging(
                handlers=[
                    sergeant.logging.logstash.LogstashHandler(
                        host='localhost',
                        port=9999,
                    ),
                ],
            ),
        )
        worker.init_logger()
        self.assertIsNotNone(
            obj=worker.logger,
        )
        self.assertEqual(
            first=len(worker.logger.handlers),
            second=1,
        )
        self.assertIsInstance(
            obj=worker.logger.handlers[0],
            cls=sergeant.logging.logstash.LogstashHandler,
        )

    def test_init_task_queue(
        self,
    ):
        worker = sergeant.worker.Worker()
        worker.config = sergeant.config.WorkerConfig(
            name='some_worker',
            connector=sergeant.config.Connector(
                type='redis',
                params={
                    'host': 'localhost',
                    'port': 6379,
                    'password': None,
                    'database': 0,
                },
            ),
        )
        worker.init_task_queue()
        self.assertIsInstance(
            obj=worker.task_queue,
            cls=sergeant.task_queue.TaskQueue,
        )
        self.assertIsInstance(
            obj=worker.task_queue.connector,
            cls=sergeant.connector.redis.Connector,
        )
        self.assertIsNone(
            obj=worker.task_queue.encoder.compressor,
        )
        self.assertIsInstance(
            obj=worker.task_queue.encoder.serializer,
            cls=sergeant.encoder.serializer.pickle.Serializer,
        )

        worker.config = sergeant.config.WorkerConfig(
            name='some_worker',
            connector=sergeant.config.Connector(
                type='redis_cluster',
                params={
                    'nodes': [
                        {
                            'host': '127.0.0.1',
                            'port': 6379,
                            'password': None,
                            'database': 0,
                        },
                        {
                            'host': '127.0.0.1',
                            'port': 6380,
                            'password': None,
                            'database': 0,
                        },
                    ],
                },
            ),
        )
        worker.init_task_queue()
        self.assertIsInstance(
            obj=worker.task_queue,
            cls=sergeant.task_queue.TaskQueue,
        )
        self.assertIsInstance(
            obj=worker.task_queue.connector,
            cls=sergeant.connector.redis_cluster.Connector,
        )
        self.assertIsNone(
            obj=worker.task_queue.encoder.compressor,
        )
        self.assertIsInstance(
            obj=worker.task_queue.encoder.serializer,
            cls=sergeant.encoder.serializer.pickle.Serializer,
        )

        worker.config = sergeant.config.WorkerConfig(
            name='some_worker',
            connector=sergeant.config.Connector(
                type='mongo',
                params={
                    'mongodb_uri': 'mongodb://localhost:27017/',
                },
            ),
        )
        worker.init_task_queue()
        self.assertIsInstance(
            obj=worker.task_queue,
            cls=sergeant.task_queue.TaskQueue,
        )
        self.assertIsInstance(
            obj=worker.task_queue.connector,
            cls=sergeant.connector.mongo.Connector,
        )
        self.assertIsNone(
            obj=worker.task_queue.encoder.compressor,
        )
        self.assertIsInstance(
            obj=worker.task_queue.encoder.serializer,
            cls=sergeant.encoder.serializer.pickle.Serializer,
        )

        compressor_names = list(sergeant.encoder.compressor.__compressors__.keys())
        compressor_names.append(None)
        serializer_names = sergeant.encoder.serializer.__serializers__.keys()
        for compressor_name in compressor_names:
            for serializer_name in serializer_names:
                worker.config = sergeant.config.WorkerConfig(
                    name='some_worker',
                    connector=sergeant.config.Connector(
                        type='redis',
                        params={
                            'host': 'localhost',
                            'port': 6379,
                            'password': None,
                            'database': 0,
                        },
                    ),
                    encoder=sergeant.config.Encoder(
                        compressor=compressor_name,
                        serializer=serializer_name,
                    ),
                )
                worker.init_task_queue()
                self.assertIsInstance(
                    obj=worker.task_queue,
                    cls=sergeant.task_queue.TaskQueue,
                )
                self.assertIsInstance(
                    obj=worker.task_queue.connector,
                    cls=sergeant.connector.redis.Connector,
                )
                if compressor_name:
                    self.assertEqual(
                        first=worker.task_queue.encoder.compressor.name,
                        second=compressor_name,
                    )
                else:
                    self.assertIsNone(
                        obj=worker.task_queue.encoder.compressor,
                    )
                self.assertEqual(
                    first=worker.task_queue.encoder.serializer.name,
                    second=serializer_name,
                )

    def test_init_executor(
        self,
    ):
        worker = sergeant.worker.Worker()
        worker.config = sergeant.config.WorkerConfig(
            name='some_worker',
            connector=sergeant.config.Connector(
                type='',
                params={},
            ),
        )
        worker.init_executor()
        self.assertIsInstance(
            obj=worker.executor_obj,
            cls=sergeant.executor.serial.SerialExecutor,
        )

        worker.config = sergeant.config.WorkerConfig(
            name='some_worker',
            connector=sergeant.config.Connector(
                type='',
                params={},
            ),
            executor=sergeant.config.Executor(
                type='serial',
            )
        )
        worker.init_executor()
        self.assertIsInstance(
            obj=worker.executor_obj,
            cls=sergeant.executor.serial.SerialExecutor,
        )

        worker.config = sergeant.config.WorkerConfig(
            name='some_worker',
            connector=sergeant.config.Connector(
                type='',
                params={},
            ),
            executor=sergeant.config.Executor(
                type='threaded',
                number_of_threads=1,
            )
        )
        worker.init_executor()
        self.assertIsInstance(
            obj=worker.executor_obj,
            cls=sergeant.executor.threaded.ThreadedExecutor,
        )

    def test_task_queue_actions(
        self,
    ):
        worker = sergeant.worker.Worker()
        worker.config = sergeant.config.WorkerConfig(
            name='some_worker',
            connector=sergeant.config.Connector(
                type='redis',
                params={
                    'host': 'localhost',
                    'port': 6379,
                    'password': None,
                    'database': 0,
                },
            ),
        )
        worker.init_task_queue()

        worker.purge_tasks()
        self.assertEqual(
            first=worker.number_of_enqueued_tasks(),
            second=0,
        )
        worker.apply_async_one(
            kwargs={
                'task': 1,
            },
        )
        self.assertEqual(
            first=worker.number_of_enqueued_tasks(),
            second=1,
        )
        worker.apply_async_many(
            kwargs_list=[
                {
                    'task': 2,
                },
                {
                    'task': 3,
                },
                {
                    'task': 4,
                },
            ],
        )
        self.assertEqual(
            first=worker.number_of_enqueued_tasks(),
            second=4,
        )
        tasks = list(
            worker.get_next_tasks(
                number_of_tasks=1,
            )
        )
        self.assertEqual(
            first=tasks[0]['kwargs']['task'],
            second=1,
        )
        self.assertEqual(
            first=tasks[0]['name'],
            second=worker.config.name,
        )
        worker.purge_tasks()
        self.assertEqual(
            first=worker.number_of_enqueued_tasks(),
            second=0,
        )

        worker.purge_tasks(
            task_name='other_worker',
        )
        self.assertEqual(
            first=worker.number_of_enqueued_tasks(
                task_name='other_worker',
            ),
            second=0,
        )
        worker.apply_async_one(
            task_name='other_worker',
            kwargs={
                'task': 1,
            },
        )
        self.assertEqual(
            first=worker.number_of_enqueued_tasks(
                task_name='other_worker',
            ),
            second=1,
        )
        worker.apply_async_many(
            task_name='other_worker',
            kwargs_list=[
                {
                    'task': 2,
                },
                {
                    'task': 3,
                },
                {
                    'task': 4,
                },
            ],
        )
        self.assertEqual(
            first=worker.number_of_enqueued_tasks(
                task_name='other_worker',
            ),
            second=4,
        )
        tasks = list(
            worker.get_next_tasks(
                task_name='other_worker',
                number_of_tasks=1,
            )
        )
        self.assertEqual(
            first=tasks[0]['kwargs']['task'],
            second=1,
        )
        self.assertEqual(
            first=tasks[0]['name'],
            second='other_worker',
        )
        worker.purge_tasks(
            task_name='other_worker',
        )
        self.assertEqual(
            first=worker.number_of_enqueued_tasks(
                task_name='other_worker',
            ),
            second=0,
        )

    def test_iterate_tasks(
        self,
    ):
        worker = sergeant.worker.Worker()
        worker.config = sergeant.config.WorkerConfig(
            name='some_worker',
            connector=sergeant.config.Connector(
                type='redis',
                params={
                    'host': 'localhost',
                    'port': 6379,
                    'password': None,
                    'database': 0,
                },
            ),
        )

        worker.iterate_tasks_forever = unittest.mock.MagicMock()
        worker.iterate_tasks_until_max_tasks = unittest.mock.MagicMock()
        worker.config.max_tasks_per_run = 0
        list(worker.iterate_tasks())
        self.assertEqual(
            first=worker.iterate_tasks_forever.call_count,
            second=1,
        )
        self.assertEqual(
            first=worker.iterate_tasks_until_max_tasks.call_count,
            second=0,
        )

        worker.iterate_tasks_forever = unittest.mock.MagicMock()
        worker.iterate_tasks_until_max_tasks = unittest.mock.MagicMock()
        worker.config.max_tasks_per_run = 1
        list(worker.iterate_tasks())
        self.assertEqual(
            first=worker.iterate_tasks_forever.call_count,
            second=0,
        )
        self.assertEqual(
            first=worker.iterate_tasks_until_max_tasks.call_count,
            second=1,
        )

    def test_iterate_tasks_until_max_tasks(
        self,
    ):
        worker = sergeant.worker.Worker()
        worker.config = sergeant.config.WorkerConfig(
            name='some_worker',
            connector=sergeant.config.Connector(
                type='redis',
                params={
                    'host': 'localhost',
                    'port': 6379,
                    'password': None,
                    'database': 0,
                },
            ),
        )

        worker.config.max_tasks_per_run = 10
        worker.get_next_tasks = unittest.mock.MagicMock(
            return_value=[
                {
                    'name': 'some_name',
                },
            ],
        )
        for task in worker.iterate_tasks_until_max_tasks():
            self.assertEqual(
                first=task,
                second={
                    'name': 'some_name',
                },
            )
        self.assertEqual(
            first=worker.get_next_tasks.call_count,
            second=10,
        )

        worker.config.max_tasks_per_run = 10
        worker.get_next_tasks = unittest.mock.MagicMock(
            return_value=[
                {
                    'name': 'some_name',
                },
            ] * 5,
        )
        for task in worker.iterate_tasks_until_max_tasks():
            self.assertEqual(
                first=task,
                second={
                    'name': 'some_name',
                },
            )
        self.assertEqual(
            first=worker.get_next_tasks.call_count,
            second=2,
        )

    def test_retry(
        self,
    ):
        worker = sergeant.worker.Worker()
        worker.config = sergeant.config.WorkerConfig(
            name='some_worker',
            connector=sergeant.config.Connector(
                type='redis',
                params={
                    'host': 'localhost',
                    'port': 6379,
                    'password': None,
                    'database': 0,
                },
            ),
            max_retries=3,
        )
        worker.init_task_queue()

        task = worker.task_queue.craft_task(
            task_name=worker.config.name,
            kwargs={},
        )
        worker.task_queue.apply_async_one = unittest.mock.MagicMock()

        with self.assertRaises(
            expected_exception=sergeant.worker.WorkerRetry,
        ):
            worker.retry(
                task=task,
            )
        self.assertEqual(
            first=worker.task_queue.apply_async_one.call_count,
            second=1,
        )

        with self.assertRaises(
            expected_exception=sergeant.worker.WorkerRetry,
        ):
            worker.retry(
                task=worker.task_queue.apply_async_one.call_args[1]['task'],
            )
        self.assertEqual(
            first=worker.task_queue.apply_async_one.call_count,
            second=2,
        )

        with self.assertRaises(
            expected_exception=sergeant.worker.WorkerRetry,
        ):
            worker.retry(
                task=worker.task_queue.apply_async_one.call_args[1]['task'],
            )
        self.assertEqual(
            first=worker.task_queue.apply_async_one.call_count,
            second=3,
        )

        with self.assertRaises(
            expected_exception=sergeant.worker.WorkerMaxRetries,
        ):
            worker.retry(
                task=worker.task_queue.apply_async_one.call_args[1]['task'],
            )
        self.assertEqual(
            first=worker.task_queue.apply_async_one.call_count,
            second=3,
        )

        worker.config.max_retries = 0
        for i in range(100):
            with self.assertRaises(
                expected_exception=sergeant.worker.WorkerRetry,
            ):
                worker.retry(
                    task=worker.task_queue.apply_async_one.call_args[1]['task'],
                )
        self.assertEqual(
            first=worker.task_queue.apply_async_one.call_count,
            second=103,
        )

    def test_requeue(
        self,
    ):
        worker = sergeant.worker.Worker()
        worker.config = sergeant.config.WorkerConfig(
            name='some_worker',
            connector=sergeant.config.Connector(
                type='redis',
                params={
                    'host': 'localhost',
                    'port': 6379,
                    'password': None,
                    'database': 0,
                },
            ),
            max_retries=3,
        )
        worker.init_task_queue()

        task = worker.task_queue.craft_task(
            task_name=worker.config.name,
            kwargs={},
        )
        worker.task_queue.apply_async_one = unittest.mock.MagicMock()

        with self.assertRaises(
            expected_exception=sergeant.worker.WorkerRequeue,
        ):
            worker.requeue(
                task=task,
            )

        self.assertEqual(
            first=worker.task_queue.apply_async_one.call_count,
            second=1,
        )

    def test_on_success(
        self,
    ):
        worker = sergeant.worker.Worker()
        worker.config = sergeant.config.WorkerConfig(
            name='some_worker',
            connector=sergeant.config.Connector(
                type='redis',
                params={
                    'host': 'localhost',
                    'port': 6379,
                    'password': None,
                    'database': 0,
                },
            ),
            max_retries=3,
            logging=sergeant.config.Logging(
                events=sergeant.config.LoggingEvents(
                    on_success=False,
                ),
            ),
        )
        worker.init_task_queue()

        task = worker.task_queue.craft_task(
            task_name=worker.config.name,
            kwargs={},
        )
        worker._on_success(
            task=task,
            returned_value=True,
        )

        worker.on_success = unittest.mock.MagicMock()
        worker._on_success(
            task=task,
            returned_value=True,
        )
        worker.on_success.assert_called_once()
        worker.config.logging.events.on_success = True
        worker.logger = unittest.mock.MagicMock()
        worker._on_success(
            task=task,
            returned_value=True,
        )
        worker.logger.info.assert_called_once_with(
            msg='task has finished successfully',
            extra={
                'task': task,
            },
        )

        worker.on_success.side_effect = Exception('exception message')
        worker._on_success(
            task=task,
            returned_value=True,
        )
        worker.logger.error.assert_called_once_with(
            msg='on_success handler has failed: exception message',
            extra={
                'task': task,
            },
        )

    def test_on_failure(
        self,
    ):
        worker = sergeant.worker.Worker()
        worker.config = sergeant.config.WorkerConfig(
            name='some_worker',
            connector=sergeant.config.Connector(
                type='redis',
                params={
                    'host': 'localhost',
                    'port': 6379,
                    'password': None,
                    'database': 0,
                },
            ),
            max_retries=3,
            logging=sergeant.config.Logging(
                events=sergeant.config.LoggingEvents(
                    on_failure=False,
                ),
            ),
        )
        worker.init_task_queue()

        task = worker.task_queue.craft_task(
            task_name=worker.config.name,
            kwargs={},
        )
        worker._on_failure(
            task=task,
            exception=Exception('test_exception'),
        )

        worker.on_failure = unittest.mock.MagicMock()
        worker._on_failure(
            task=task,
            exception=Exception('test_exception'),
        )
        worker.on_failure.assert_called_once()
        worker.config.logging.events.on_failure = True
        worker.logger = unittest.mock.MagicMock()
        worker._on_failure(
            task=task,
            exception=Exception('test_exception'),
        )
        worker.logger.error.assert_called_once_with(
            msg='task has failed',
            extra={
                'task': task,
            },
        )

        worker.on_failure.side_effect = Exception('exception message')
        worker._on_failure(
            task=task,
            exception=Exception('test_exception'),
        )
        worker.logger.error.assert_called_with(
            msg='on_failure handler has failed: exception message',
            extra={
                'task': task,
            },
        )

    def test_on_timeout(
        self,
    ):
        worker = sergeant.worker.Worker()
        worker.config = sergeant.config.WorkerConfig(
            name='some_worker',
            connector=sergeant.config.Connector(
                type='redis',
                params={
                    'host': 'localhost',
                    'port': 6379,
                    'password': None,
                    'database': 0,
                },
            ),
            max_retries=3,
            logging=sergeant.config.Logging(
                events=sergeant.config.LoggingEvents(
                    on_timeout=False,
                ),
            ),
        )
        worker.init_task_queue()

        task = worker.task_queue.craft_task(
            task_name=worker.config.name,
            kwargs={},
        )
        worker._on_timeout(
            task=task,
        )

        worker.on_timeout = unittest.mock.MagicMock()
        worker._on_timeout(
            task=task,
        )
        worker.on_timeout.assert_called_once()
        worker.config.logging.events.on_timeout = True
        worker.logger = unittest.mock.MagicMock()
        worker._on_timeout(
            task=task,
        )
        worker.logger.error.assert_called_once_with(
            msg='task has timedout',
            extra={
                'task': task,
            },
        )

        worker.on_timeout.side_effect = Exception('exception message')
        worker._on_timeout(
            task=task,
        )
        worker.logger.error.assert_called_with(
            msg='on_timeout handler has failed: exception message',
            extra={
                'task': task,
            },
        )

    def test_on_retry(
        self,
    ):
        worker = sergeant.worker.Worker()
        worker.config = sergeant.config.WorkerConfig(
            name='some_worker',
            connector=sergeant.config.Connector(
                type='redis',
                params={
                    'host': 'localhost',
                    'port': 6379,
                    'password': None,
                    'database': 0,
                },
            ),
            max_retries=3,
            logging=sergeant.config.Logging(
                events=sergeant.config.LoggingEvents(
                    on_retry=False,
                ),
            ),
        )
        worker.init_task_queue()

        task = worker.task_queue.craft_task(
            task_name=worker.config.name,
            kwargs={},
        )
        worker._on_retry(
            task=task,
        )

        worker.on_retry = unittest.mock.MagicMock()
        worker._on_retry(
            task=task,
        )
        worker.on_retry.assert_called_once()
        worker.config.logging.events.on_retry = True
        worker.logger = unittest.mock.MagicMock()
        worker._on_retry(
            task=task,
        )
        worker.logger.info.assert_called_once_with(
            msg='task has retried',
            extra={
                'task': task,
            },
        )

        worker.on_retry.side_effect = Exception('exception message')
        worker._on_retry(
            task=task,
        )
        worker.logger.error.assert_called_once_with(
            msg='on_retry handler has failed: exception message',
            extra={
                'task': task,
            },
        )

    def test_on_max_retries(
        self,
    ):
        worker = sergeant.worker.Worker()
        worker.config = sergeant.config.WorkerConfig(
            name='some_worker',
            connector=sergeant.config.Connector(
                type='redis',
                params={
                    'host': 'localhost',
                    'port': 6379,
                    'password': None,
                    'database': 0,
                },
            ),
            max_retries=3,
            logging=sergeant.config.Logging(
                events=sergeant.config.LoggingEvents(
                    on_max_retries=False,
                ),
            ),
        )
        worker.init_task_queue()

        task = worker.task_queue.craft_task(
            task_name=worker.config.name,
            kwargs={},
        )
        worker._on_max_retries(
            task=task,
        )

        worker.on_max_retries = unittest.mock.MagicMock()
        worker._on_max_retries(
            task=task,
        )
        worker.on_max_retries.assert_called_once()
        worker.config.logging.events.on_max_retries = True
        worker.logger = unittest.mock.MagicMock()
        worker._on_max_retries(
            task=task,
        )
        worker.logger.error.assert_called_once_with(
            msg='task has reached max retries',
            extra={
                'task': task,
            },
        )

        worker.on_max_retries.side_effect = Exception('exception message')
        worker._on_max_retries(
            task=task,
        )
        worker.logger.error.assert_called_with(
            msg='on_max_retries handler has failed: exception message',
            extra={
                'task': task,
            },
        )

    def test_on_requeue(
        self,
    ):
        worker = sergeant.worker.Worker()
        worker.config = sergeant.config.WorkerConfig(
            name='some_worker',
            connector=sergeant.config.Connector(
                type='redis',
                params={
                    'host': 'localhost',
                    'port': 6379,
                    'password': None,
                    'database': 0,
                },
            ),
            max_retries=3,
            logging=sergeant.config.Logging(
                events=sergeant.config.LoggingEvents(
                    on_requeue=False,
                ),
            ),
        )
        worker.init_task_queue()

        task = worker.task_queue.craft_task(
            task_name=worker.config.name,
            kwargs={},
        )
        worker._on_requeue(
            task=task,
        )

        worker.on_requeue = unittest.mock.MagicMock()
        worker._on_requeue(
            task=task,
        )
        worker.on_requeue.assert_called_once()
        worker.config.logging.events.on_requeue = True
        worker.logger = unittest.mock.MagicMock()
        worker._on_requeue(
            task=task,
        )
        worker.logger.info.assert_called_once_with(
            msg='task has requeued',
            extra={
                'task': task,
            },
        )

        worker.on_requeue.side_effect = Exception('exception message')
        worker._on_requeue(
            task=task,
        )
        worker.logger.error.assert_called_once_with(
            msg='on_requeue handler has failed: exception message',
            extra={
                'task': task,
            },
        )



import unittest

import sergeant.encoder


class EncoderTestCase(
    unittest.TestCase,
):
    obj_to_encode = {
        'a': 1.3,
        'b': [
            1,
        ],
        'c': {
            '1': {
                'a': True,
                'b': False,
                'c': -1,
            },
            '2': None,
        },
        'd': 'unicode string \u00AE',
        'e': {
            '1': 1,
            '2': 2,
        },
    }

    def test_encoders(
        self,
    ):
        compressor_names = list(sergeant.encoder.compressor.__compressors__.keys())
        compressor_names.append(None)
        serializer_names = sergeant.encoder.serializer.__serializers__.keys()
        for compressor_name in compressor_names:
            for serializer_name in serializer_names:
                encoder_obj = sergeant.encoder.encoder.Encoder(
                    compressor_name=compressor_name,
                    serializer_name=serializer_name,
                )

                encoded = encoder_obj.encode(
                    data=self.obj_to_encode,
                )
                decoded = encoder_obj.decode(
                    data=encoded,
                )

                self.assertEqual(
                    first=decoded,
                    second=self.obj_to_encode,
                )



import unittest
import unittest.mock

import sergeant.connector


class ConnectorTestCase:
    test_queue_name = b'test_queue_name'
    test_queue_item = b'test_queue_item'
    test_queue_items = [
        b'test_queue_item_1',
        b'test_queue_item_2',
        b'test_queue_item_3',
        b'test_queue_item_4',
        b'test_queue_item_5',
        b'test_queue_item_6',
        b'test_queue_item_7',
        b'test_queue_item_8',
        b'test_queue_item_9',
        b'test_queue_item_10',
    ]

    test_key_name = b'test_key'
    test_key_value = b'test_value'

    def test_key(
        self,
    ):
        self.connector.key_delete(
            key=self.test_key_name,
        )
        key_value = self.connector.key_get(
            key=self.test_key_name,
        )
        self.assertIsNone(
            obj=key_value,
        )

        key_is_new = self.connector.key_set(
            key=self.test_key_name,
            value=self.test_key_value
        )
        self.assertTrue(
            expr=key_is_new,
        )

        key_is_new = self.connector.key_set(
            key=self.test_key_name,
            value=self.test_key_value
        )
        self.assertFalse(
            expr=key_is_new,
        )

        key_value = self.connector.key_get(
            key=self.test_key_name,
        )
        self.assertEqual(
            first=key_value,
            second=self.test_key_value,
        )

        self.connector.key_delete(
            key=self.test_key_name,
        )
        key_value = self.connector.key_get(
            key=self.test_key_name,
        )
        self.assertIsNone(
            obj=key_value,
        )

        key_is_new = self.connector.key_set(
            key=self.test_key_name,
            value=self.test_key_value
        )
        self.assertTrue(
            expr=key_is_new,
        )
        self.connector.key_delete(
            key=self.test_key_name,
        )

    def test_queue(
        self,
    ):
        self.connector.queue_delete(
            queue_name=self.test_queue_name,
        )
        queue_length = self.connector.queue_length(
            queue_name=self.test_queue_name,
        )
        self.assertEqual(
            first=queue_length,
            second=0,
        )

        self.connector.queue_push(
            queue_name=self.test_queue_name,
            item=self.test_queue_item,
            priority='NORMAL',
        )
        queue_length = self.connector.queue_length(
            queue_name=self.test_queue_name,
        )
        self.assertEqual(
            first=queue_length,
            second=1,
        )
        self.connector.queue_delete(
            queue_name=self.test_queue_name,
        )
        queue_length = self.connector.queue_length(
            queue_name=self.test_queue_name,
        )
        self.assertEqual(
            first=queue_length,
            second=0,
        )

        self.connector.queue_push(
            queue_name=self.test_queue_name,
            item=self.test_queue_item,
            priority='NORMAL',
        )
        item = self.connector.queue_pop(
            queue_name=self.test_queue_name,
        )
        self.assertEqual(
            first=item,
            second=self.test_queue_item,
        )

        for item in self.test_queue_items:
            self.connector.queue_push(
                queue_name=self.test_queue_name,
                item=item,
                priority='NORMAL',
            )

        items = []
        for i in range(len(self.test_queue_items)):
            item = self.connector.queue_pop(
                queue_name=self.test_queue_name,
            )
            items.append(item)

        self.assertEqual(
            first=items,
            second=self.test_queue_items,
        )

        self.connector.queue_push_bulk(
            queue_name=self.test_queue_name,
            items=self.test_queue_items,
            priority='NORMAL'
        )
        items = []
        for i in range(len(self.test_queue_items)):
            item = self.connector.queue_pop(
                queue_name=self.test_queue_name,
            )
            items.append(item)

        self.assertEqual(
            first=items,
            second=self.test_queue_items,
        )

        self.connector.queue_push_bulk(
            queue_name=self.test_queue_name,
            items=self.test_queue_items,
            priority='NORMAL'
        )
        items = self.connector.queue_pop_bulk(
            queue_name=self.test_queue_name,
            number_of_items=len(self.test_queue_items),
        )
        self.assertEqual(
            first=items,
            second=self.test_queue_items,
        )

        self.connector.queue_push_bulk(
            queue_name=self.test_queue_name,
            items=self.test_queue_items * 1000,
            priority='NORMAL'
        )
        items = self.connector.queue_pop_bulk(
            queue_name=self.test_queue_name,
            number_of_items=len(self.test_queue_items) * 1000,
        )
        self.assertEqual(
            first=items,
            second=self.test_queue_items * 1000,
        )

        self.connector.queue_push_bulk(
            queue_name=self.test_queue_name,
            items=self.test_queue_items * 1000,
            priority='NORMAL'
        )
        queue_length = self.connector.queue_length(
            queue_name=self.test_queue_name,
        )
        self.assertEqual(
            first=queue_length,
            second=len(self.test_queue_items) * 1000,
        )

        self.connector.queue_delete(
            queue_name=self.test_queue_name,
        )
        queue_length = self.connector.queue_length(
            queue_name=self.test_queue_name,
        )
        self.assertEqual(
            first=queue_length,
            second=0,
        )

        item = self.connector.queue_pop(
            queue_name=self.test_queue_name,
        )
        self.assertIsNone(
            obj=item,
        )

        items = self.connector.queue_pop_bulk(
            queue_name=self.test_queue_name,
            number_of_items=1,
        )
        self.assertEqual(
            first=items,
            second=[],
        )

    def test_queue_priorities(
        self,
    ):
        self.connector.queue_delete(
            queue_name=self.test_queue_name,
        )
        self.connector.queue_push(
            queue_name=self.test_queue_name,
            item=self.test_queue_item + b'1',
            priority='NORMAL',
        )
        self.connector.queue_push(
            queue_name=self.test_queue_name,
            item=self.test_queue_item + b'1',
            priority='NORMAL',
        )
        self.connector.queue_push(
            queue_name=self.test_queue_name,
            item=self.test_queue_item + b'2',
            priority='HIGH',
        )
        self.connector.queue_push(
            queue_name=self.test_queue_name,
            item=self.test_queue_item + b'2',
            priority='HIGH',
        )
        item = self.connector.queue_pop(
            queue_name=self.test_queue_name,
        )
        self.assertEqual(
            first=item,
            second=self.test_queue_item + b'2',
        )
        item = self.connector.queue_pop(
            queue_name=self.test_queue_name,
        )
        self.assertEqual(
            first=item,
            second=self.test_queue_item + b'2',
        )
        item = self.connector.queue_pop(
            queue_name=self.test_queue_name,
        )
        self.assertEqual(
            first=item,
            second=self.test_queue_item + b'1',
        )
        item = self.connector.queue_pop(
            queue_name=self.test_queue_name,
        )
        self.assertEqual(
            first=item,
            second=self.test_queue_item + b'1',
        )

        self.connector.queue_push_bulk(
            queue_name=self.test_queue_name,
            items=[
                self.test_queue_item + b'1',
            ] * 50,
            priority='NORMAL',
        )
        self.connector.queue_push_bulk(
            queue_name=self.test_queue_name,
            items=[
                self.test_queue_item + b'1',
            ] * 50,
            priority='NORMAL',
        )
        self.connector.queue_push_bulk(
            queue_name=self.test_queue_name,
            items=[
                self.test_queue_item + b'2',
            ] * 50,
            priority='HIGH',
        )
        self.connector.queue_push_bulk(
            queue_name=self.test_queue_name,
            items=[
                self.test_queue_item + b'2',
            ] * 50,
            priority='HIGH',
        )
        items = self.connector.queue_pop_bulk(
            queue_name=self.test_queue_name,
            number_of_items=50
        )
        self.assertEqual(
            first=items,
            second=[
                self.test_queue_item + b'2',
            ] * 50,
        )
        items = self.connector.queue_pop_bulk(
            queue_name=self.test_queue_name,
            number_of_items=50
        )
        self.assertEqual(
            first=items,
            second=[
                self.test_queue_item + b'2',
            ] * 50,
        )
        items = self.connector.queue_pop_bulk(
            queue_name=self.test_queue_name,
            number_of_items=50
        )
        self.assertEqual(
            first=items,
            second=[
                self.test_queue_item + b'1',
            ] * 50,
        )
        items = self.connector.queue_pop_bulk(
            queue_name=self.test_queue_name,
            number_of_items=50
        )
        self.assertEqual(
            first=items,
            second=[
                self.test_queue_item + b'1',
            ] * 50,
        )


class RedisConnectorTestCase(
    ConnectorTestCase,
    unittest.TestCase,
):
    def setUp(
        self,
    ):
        self.connector = sergeant.connector.redis.Connector(
            host='127.0.0.1',
            port=6379,
            password=None,
            database=0,
        )


class RedisClusterSingleServerConnectorTestCase(
    ConnectorTestCase,
    unittest.TestCase,
):
    def setUp(
        self,
    ):
        self.connector = sergeant.connector.redis_cluster.Connector(
            nodes=[
                {
                    'host': '127.0.0.1',
                    'port': 6379,
                    'password': None,
                    'database': 0,
                },
            ]
        )


class RedisClusterMultipleServersConnectorTestCase(
    ConnectorTestCase,
    unittest.TestCase,
):
    def setUp(
        self,
    ):
        self.connector = sergeant.connector.redis_cluster.Connector(
            nodes=[
                {
                    'host': '127.0.0.1',
                    'port': 6379,
                    'password': None,
                    'database': 0,
                },
                {
                    'host': '127.0.0.1',
                    'port': 6380,
                    'password': None,
                    'database': 0,
                },
            ]
        )


class MongoConnectorTestCase(
    ConnectorTestCase,
    unittest.TestCase,
):
    def setUp(
        self,
    ):
        self.connector = sergeant.connector.mongo.Connector(
            mongodb_uri='mongodb://localhost:27017/',
        )



import setuptools


setuptools.setup(
    name='sergeant',
    version='0.6.0',
    author='Gal Ben David',
    author_email='gal@intsights.com',
    url='https://github.com/Intsights/sergeant',
    project_urls={
        'Source': 'https://github.com/Intsights/sergeant',
    },
    license='MIT',
    description='Fast, Safe & Simple Asynchronous Task Queues Written In Pure Python',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    keywords='tasks worker queue redis async',
    python_requires='>=3.7',
    zip_safe=False,
    install_requires=[
        'msgpack',
        'orjson',
        'psutil',
        'pymongo',
        'redis',
    ],
    tests_require=[],
    package_data={},
    packages=setuptools.find_packages(),
)



import importlib
import traceback
import sys
import multiprocessing.connection
import argparse


def main() -> int:
    parser = argparse.ArgumentParser(
        description='Sergeant Supervisor',
    )
    parser.add_argument(
        '--child-pipe',
        help='Pipe fileno to return the potential exception',
        type=int,
        required=True,
        dest='child_pipe',
    )
    parser.add_argument(
        '--worker-class',
        help='Class name of the worker to spawn',
        type=str,
        required=True,
        dest='worker_class',
    )
    parser.add_argument(
        '--worker-module',
        help='Module of the worker class',
        type=str,
        required=True,
        dest='worker_module',
    )
    args = parser.parse_args()
    pipe_obj = multiprocessing.connection.Connection(
        handle=args.child_pipe,
    )

    try:
        worker_module = importlib.import_module(
            name=args.worker_module,
        )
    except ModuleNotFoundError:
        return 2

    try:
        worker_class = getattr(worker_module, args.worker_class)
    except AttributeError:
        return 3

    try:
        worker_obj = worker_class()
        worker_obj.init_worker()
        summary = worker_obj.work_loop()
        pipe_obj.send(summary)

        return 0
    except KeyboardInterrupt:
        return 0
    except Exception as exception:
        pipe_obj.send(
            {
                'exception': repr(exception),
                'traceback': traceback.format_exc(),
            }
        )

        return 1
    finally:
        try:
            pipe_obj.close()
        except Exception:
            pass


if __name__ == '__main__':
    sys.exit(main())



import logging
import dataclasses
import typing


@dataclasses.dataclass
class LoggingEvents:
    on_success: bool = False
    on_failure: bool = True
    on_timeout: bool = True
    on_retry: bool = True
    on_max_retries: bool = True
    on_requeue: bool = True


@dataclasses.dataclass
class Logging:
    level: int = logging.ERROR
    log_to_stdout: bool = False
    events: LoggingEvents = dataclasses.field(
        default_factory=LoggingEvents,
    )
    handlers: typing.List[logging.Handler] = dataclasses.field(
        default_factory=list,
    )


@dataclasses.dataclass
class Encoder:
    compressor: typing.Optional[str] = None
    serializer: str = 'pickle'


@dataclasses.dataclass
class Connector:
    type: str
    params: typing.Dict[str, typing.Any]


@dataclasses.dataclass
class Executor:
    type: str = 'serial'
    number_of_threads: int = 1


@dataclasses.dataclass
class Timeouts:
    soft_timeout: float = 0.0
    hard_timeout: float = 0.0
    critical_timeout: float = 0.0


@dataclasses.dataclass
class WorkerConfig:
    name: str
    connector: Connector
    max_tasks_per_run: int = 0
    max_retries: int = 0
    tasks_per_transaction: int = 1
    encoder: Encoder = dataclasses.field(
        default_factory=Encoder,
    )
    executor: Executor = dataclasses.field(
        default_factory=Executor,
    )
    timeouts: Timeouts = dataclasses.field(
        default_factory=Timeouts,
    )
    logging: Logging = dataclasses.field(
        default_factory=Logging,
    )



import redis
import typing


class Connector:
    name = 'redis'

    def __init__(
        self,
        host: str,
        port: int,
        password: typing.Optional[str],
        database: int,
    ) -> None:
        self.host = host
        self.port = port
        self.password = password
        self.database = database

        self.connection = redis.Redis(
            host=self.host,
            port=self.port,
            password=self.password,
            db=self.database,
            retry_on_timeout=True,
            socket_keepalive=True,
            socket_connect_timeout=10,
            socket_timeout=60,
        )

    def key_set(
        self,
        key: str,
        value: bytes,
    ) -> bool:
        is_new = self.connection.set(
            name=key,
            value=value,
            nx=True,
        )

        return is_new is True

    def key_get(
        self,
        key: str,
    ) -> bytes:
        return self.connection.get(
            name=key,
        )

    def key_delete(
        self,
        key: str,
    ) -> bool:
        return self.connection.delete(key) > 0

    def queue_pop(
        self,
        queue_name: str,
    ) -> typing.Optional[bytes]:
        value = self.connection.lpop(
            name=queue_name,
        )

        if value is None:
            return None
        else:
            return value

    def queue_pop_bulk(
        self,
        queue_name: str,
        number_of_items: int,
    ) -> typing.List[bytes]:
        pipeline = self.connection.pipeline()

        pipeline.lrange(queue_name, 0, number_of_items - 1)
        pipeline.ltrim(queue_name, number_of_items, -1)

        value = pipeline.execute()

        return value[0]

    def queue_push(
        self,
        queue_name: str,
        item: bytes,
        priority: str = 'NORMAL',
    ) -> bool:
        if priority == 'HIGH':
            self.connection.lpush(queue_name, item)
        else:
            self.connection.rpush(queue_name, item)

        return True

    def queue_push_bulk(
        self,
        queue_name: str,
        items: typing.Iterable[bytes],
        priority: str = 'NORMAL',
    ) -> bool:
        if priority == 'HIGH':
            self.connection.lpush(queue_name, *items)
        else:
            self.connection.rpush(queue_name, *items)

        return True

    def queue_length(
        self,
        queue_name: str,
    ) -> int:
        return self.connection.llen(
            name=queue_name,
        )

    def queue_delete(
        self,
        queue_name: str,
    ) -> bool:
        return self.connection.delete(queue_name) > 0



import time
import datetime
import logging
import typing

from . import config
from . import connector
from . import encoder
from . import task_queue
from . import executor


class Worker:
    config = config.WorkerConfig(
        name='test_worker',
        connector=config.Connector(
            type='redis',
            params={},
        ),
    )

    def init_worker(
        self,
    ) -> None:
        self.init_task_queue()
        self.init_logger()
        self.init_executor()

    def init_logger(
        self,
    ) -> None:
        self.logger = logging.getLogger(
            name=self.config.name,
        )
        self.logger.propagate = False
        self.logger.setLevel(
            level=self.config.logging.level,
        )

        if self.config.logging.log_to_stdout:
            stream_handler = logging.StreamHandler()
            stream_handler.setFormatter(
                fmt=logging.Formatter(
                    fmt='%(asctime)s %(name)-12s %(process)d %(levelname)-8s %(message)s',
                    datefmt='%Y-%m-%d %H:%M:%S',
                ),
            )
            self.logger.addHandler(
                hdlr=stream_handler,
            )

        for handler in self.config.logging.handlers:
            self.logger.addHandler(
                hdlr=handler,
            )

    def init_task_queue(
        self,
    ) -> None:
        encoder_obj = encoder.encoder.Encoder(
            compressor_name=self.config.encoder.compressor,
            serializer_name=self.config.encoder.serializer,
        )
        connector_class = connector.__connectors__[self.config.connector.type]
        connector_obj = connector_class(**self.config.connector.params)
        self.task_queue = task_queue.TaskQueue(
            connector=connector_obj,
            encoder=encoder_obj,
        )

    def init_executor(
        self,
    ) -> None:
        if self.config.executor.type == 'serial':
            self.executor_obj = executor.serial.SerialExecutor(
                worker=self,
            )
        elif self.config.executor.type == 'threaded':
            self.executor_obj = executor.threaded.ThreadedExecutor(
                worker=self,
                number_of_threads=self.config.executor.number_of_threads,
            )

    def purge_tasks(
        self,
        task_name: typing.Optional[str] = None,
    ) -> bool:
        try:
            return self.task_queue.purge_tasks(
                task_name=task_name if task_name else self.config.name,
            )
        except Exception as exception:
            self.logger.error(
                msg=f'could not purge tasks: {exception}',
            )

            return False

    def number_of_enqueued_tasks(
        self,
        task_name: typing.Optional[str] = None,
    ) -> typing.Optional[int]:
        try:
            return self.task_queue.number_of_enqueued_tasks(
                task_name=task_name if task_name else self.config.name,
            )
        except Exception as exception:
            self.logger.error(
                msg=f'could not get the queue length: {exception}',
            )

            return None

    def apply_async_one(
        self,
        kwargs: typing.Dict[str, typing.Any],
        task_name: typing.Optional[str] = None,
        priority: str = 'NORMAL',
    ) -> bool:
        try:
            task = self.task_queue.craft_task(
                task_name=task_name if task_name else self.config.name,
                kwargs=kwargs,
            )

            self.task_queue.apply_async_one(
                task=task,
                priority=priority,
            )

            return True
        except Exception as exception:
            self.logger.error(
                msg=f'could not push task: {exception}',
            )

            return False

    def apply_async_many(
        self,
        kwargs_list: typing.Iterable[typing.Dict[str, typing.Any]],
        task_name: typing.Optional[str] = None,
        priority: str = 'NORMAL',
    ) -> bool:
        try:
            if task_name is None:
                task_name = self.config.name

            tasks = [
                self.task_queue.craft_task(
                    task_name=task_name,
                    kwargs=kwargs,
                )
                for kwargs in kwargs_list
            ]

            return self.task_queue.apply_async_many(
                task_name=task_name,
                tasks=tasks,
                priority=priority,
            )
        except Exception as exception:
            self.logger.error(
                msg=f'could not push tasks: {exception}',
            )

            return False

    def get_next_tasks(
        self,
        number_of_tasks: int,
        task_name: typing.Optional[str] = None,
    ) -> typing.List[typing.Dict[str, typing.Any]]:
        try:
            return self.task_queue.get_tasks(
                task_name=task_name if task_name else self.config.name,
                number_of_tasks=number_of_tasks,
            )
        except Exception as exception:
            self.logger.error(
                msg=f'could not pull tasks: {exception}',
            )

            return []

    def iterate_tasks(
        self,
    ) -> typing.Iterable[typing.Dict[str, typing.Any]]:
        run_forever = self.config.max_tasks_per_run == 0
        if run_forever:
            yield from self.iterate_tasks_forever()
        else:
            yield from self.iterate_tasks_until_max_tasks()

    def iterate_tasks_forever(
        self,
    ) -> typing.Iterable[typing.Dict[str, typing.Any]]:
        while True:
            tasks = self.get_next_tasks(
                number_of_tasks=self.config.tasks_per_transaction,
            )
            if len(tasks) == 0:
                time.sleep(1)

                continue

            for task in tasks:
                yield task

    def iterate_tasks_until_max_tasks(
        self,
    ) -> typing.Iterable[typing.Dict[str, typing.Any]]:
        tasks_left = self.config.max_tasks_per_run
        while tasks_left > 0:
            tasks = self.get_next_tasks(
                number_of_tasks=min(
                    self.config.tasks_per_transaction,
                    tasks_left,
                ),
            )
            number_of_dequeued_tasks = len(tasks)
            if number_of_dequeued_tasks == 0:
                time.sleep(1)

                continue

            for task in tasks:
                yield task

            tasks_left -= number_of_dequeued_tasks

    def work_loop(
        self,
    ) -> typing.Dict[str, typing.Any]:
        summary = {
            'start_time': datetime.datetime.utcnow(),
            'executor_exception': None,
            'initialize_exception': None,
            'finalize_exception': None,
        }

        try:
            self.initialize()
        except Exception as exception:
            self.logger.error(
                msg=f'initialize has failed: {exception}',
            )

            summary['end_time'] = datetime.datetime.utcnow()
            summary['initialize_exception'] = exception

            return summary

        try:
            self.executor_obj.execute_tasks(
                tasks=self.iterate_tasks(),
            )
        except Exception as exception:
            self.logger.error(
                msg=f'execute_tasks has failed: {exception}',
            )

            summary['executor_exception'] = exception
        finally:
            try:
                self.finalize()
            except Exception as exception:
                self.logger.error(
                    msg=f'finalize has failed: {exception}',
                )
                summary['finalize_exception'] = exception

            summary['end_time'] = datetime.datetime.utcnow()

            return summary

    def retry(
        self,
        task: typing.Dict[str, typing.Any],
        priority: str = 'NORMAL',
    ) -> None:
        if self.config.max_retries > 0 and self.config.max_retries <= task['run_count']:
            raise WorkerMaxRetries()
        else:
            self.task_queue.retry(
                task=task,
                priority=priority,
            )

            raise WorkerRetry()

    def requeue(
        self,
        task: typing.Dict[str, typing.Any],
        priority: str = 'NORMAL',
    ) -> None:
        self.task_queue.requeue(
            task=task,
            priority=priority,
        )

        raise WorkerRequeue()

    def _on_success(
        self,
        task: typing.Dict[str, typing.Any],
        returned_value: typing.Any,
    ) -> None:
        try:
            if self.config.logging.events.on_success:
                self.logger.info(
                    msg='task has finished successfully',
                    extra={
                        'task': task,
                    },
                )

            self.on_success(
                task=task,
                returned_value=returned_value,
            )
        except Exception as exception:
            self.logger.error(
                msg=f'on_success handler has failed: {exception}',
                extra={
                    'task': task,
                },
            )

    def _on_failure(
        self,
        task: typing.Dict[str, typing.Any],
        exception: Exception,
    ) -> None:
        try:
            if self.config.logging.events.on_failure:
                self.logger.error(
                    msg='task has failed',
                    extra={
                        'task': task,
                    },
                )

            self.on_failure(
                task=task,
                exception=exception,
            )
        except Exception as exception:
            self.logger.error(
                msg=f'on_failure handler has failed: {exception}',
                extra={
                    'task': task,
                },
            )

    def _on_timeout(
        self,
        task: typing.Dict[str, typing.Any],
    ) -> None:
        try:
            if self.config.logging.events.on_timeout:
                self.logger.error(
                    msg='task has timedout',
                    extra={
                        'task': task,
                    },
                )

            self.on_timeout(
                task=task,
            )
        except Exception as exception:
            self.logger.error(
                msg=f'on_timeout handler has failed: {exception}',
                extra={
                    'task': task,
                },
            )

    def _on_retry(
        self,
        task: typing.Dict[str, typing.Any],
    ) -> None:
        try:
            if self.config.logging.events.on_retry:
                self.logger.info(
                    msg='task has retried',
                    extra={
                        'task': task,
                    },
                )

            self.on_retry(
                task=task,
            )
        except Exception as exception:
            self.logger.error(
                msg=f'on_retry handler has failed: {exception}',
                extra={
                    'task': task,
                },
            )

    def _on_max_retries(
        self,
        task: typing.Dict[str, typing.Any],
    ) -> None:
        try:
            if self.config.logging.events.on_max_retries:
                self.logger.error(
                    msg='task has reached max retries',
                    extra={
                        'task': task,
                    },
                )

            self.on_max_retries(
                task=task,
            )
        except Exception as exception:
            self.logger.error(
                msg=f'on_max_retries handler has failed: {exception}',
                extra={
                    'task': task,
                },
            )

    def _on_requeue(
        self,
        task: typing.Dict[str, typing.Any],
    ) -> None:
        try:
            if self.config.logging.events.on_requeue:
                self.logger.info(
                    msg='task has requeued',
                    extra={
                        'task': task,
                    },
                )

            self.on_requeue(
                task=task,
            )
        except Exception as exception:
            self.logger.error(
                msg=f'on_requeue handler has failed: {exception}',
                extra={
                    'task': task,
                },
            )

    def initialize(
        self,
    ) -> None:
        pass

    def finalize(
        self,
    ) -> None:
        pass

    def pre_work(
        self,
        task: typing.Dict[str, typing.Any],
    ) -> None:
        pass

    def post_work(
        self,
        task: typing.Dict[str, typing.Any],
        success: bool,
        exception: typing.Optional[Exception],
    ) -> None:
        pass

    def work(
        self,
        task: typing.Dict[str, typing.Any],
    ) -> typing.Any:
        pass

    def on_success(
        self,
        task: typing.Dict[str, typing.Any],
        returned_value: typing.Any,
    ) -> None:
        pass

    def on_failure(
        self,
        task: typing.Dict[str, typing.Any],
        exception: Exception,
    ) -> None:
        pass

    def on_timeout(
        self,
        task: typing.Dict[str, typing.Any],
    ) -> None:
        pass

    def on_retry(
        self,
        task: typing.Dict[str, typing.Any],
    ) -> None:
        pass

    def on_requeue(
        self,
        task: typing.Dict[str, typing.Any],
    ) -> None:
        pass

    def on_max_retries(
        self,
        task: typing.Dict[str, typing.Any],
    ) -> None:
        pass


class WorkerException(
    Exception,
):
    pass


class WorkerTimedout(
    WorkerException,
):
    pass


class WorkerHardTimedout(
    WorkerTimedout,
):
    pass


class WorkerSoftTimedout(
    WorkerTimedout,
):
    pass


class WorkerRequeue(
    WorkerException,
):
    pass


class WorkerRetry(
    WorkerException,
):
    pass


class WorkerMaxRetries(
    WorkerException,
):
    pass



import unittest
import datetime

import sergeant.encoder.serializer


class SerializersTestCase(
    unittest.TestCase,
):
    def test_msgpack(
        self,
    ):
        obj_to_serialize = {
            'a': 1.3,
            'b': [
                1,
            ],
            'c': {
                '1': {
                    'a': True,
                    'b': False,
                    'c': -1,
                },
                '2': None,
            },
            'd': 'unicode string \u00AE',
            'e': {
                1: 1,
                2: 2,
            },
        }

        serialized_object = sergeant.encoder.serializer.msgpack.Serializer.serialize(
            data=obj_to_serialize,
        )
        unserialized_object = sergeant.encoder.serializer.msgpack.Serializer.unserialize(
            data=serialized_object,
        )

        self.assertEqual(
            first=unserialized_object,
            second=obj_to_serialize,
        )

    def test_json(
        self,
    ):
        obj_to_serialize = {
            'a': 1.3,
            'b': [
                1,
            ],
            'c': {
                '1': {
                    'a': True,
                    'b': False,
                    'c': -1,
                },
                '2': None,
            },
            'd': 'unicode string \u00AE',
            'e': {
                '1': 1,
                '2': 2,
            },
        }

        serialized_object = sergeant.encoder.serializer.json.Serializer.serialize(
            data=obj_to_serialize,
        )
        unserialized_object = sergeant.encoder.serializer.json.Serializer.unserialize(
            data=serialized_object,
        )

        self.assertEqual(
            first=unserialized_object,
            second=obj_to_serialize,
        )

    def test_pickle(
        self,
    ):
        obj_to_serialize = {
            'a': 1.3,
            'b': [
                1,
            ],
            'c': {
                '1': {
                    'a': True,
                    'b': False,
                    'c': -1,
                },
                '2': None,
            },
            'd': 'unicode string \u00AE',
            'e': {
                1: 1,
                2: 2,
            },
            'f': datetime.datetime.utcnow(),
            'g': (
                1,
                2,
            ),
        }

        serialized_object = sergeant.encoder.serializer.pickle.Serializer.serialize(
            data=obj_to_serialize,
        )
        unserialized_object = sergeant.encoder.serializer.pickle.Serializer.unserialize(
            data=serialized_object,
        )

        self.assertEqual(
            first=unserialized_object,
            second=obj_to_serialize,
        )
